from typing import Any, ClassVar

from textual.widgets import Static

from .base_renderer import BaseToolRenderer
from .registry import register_tool_renderer


@register_tool_renderer
class FinishScanRenderer(BaseToolRenderer):
    tool_name: ClassVar[str] = "finish_scan"
    css_classes: ClassVar[list[str]] = ["tool-call", "finish-tool"]

    @classmethod
    def render(cls, tool_data: dict[str, Any]) -> Static:
        args = tool_data.get("args", {})

        content = args.get("content", "")
        success = args.get("success", True)

        header = (
            "🏁 [bold #dc2626]Finishing Scan[/]" if success else "🏁 [bold #dc2626]Scan Failed[/]"
        )

        if content:
            content_text = f"{header}\n  [bold]{cls.escape_markup(content)}[/]"
        else:
            content_text = f"{header}\n  [dim]Generating final report...[/]"

        css_classes = cls.get_css_classes("completed")
        return Static(content_text, classes=css_classes)



from .agents_graph_actions import (
    agent_finish,
    create_agent,
    send_message_to_agent,
    view_agent_graph,
    wait_for_message,
)


__all__ = [
    "agent_finish",
    "create_agent",
    "send_message_to_agent",
    "view_agent_graph",
    "wait_for_message",
]



import atexit
import os
import signal
import sys
import threading
import time
from typing import Any

from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from exaaiagnt.agents.ExaaiAgent import ExaaiAgent
from exaaiagnt.llm.config import LLMConfig
from exaaiagnt.telemetry.tracer import Tracer, set_global_tracer

from .utils import build_final_stats_text, build_live_stats_text, get_severity_color


# Clean ASCII Banner
BANNER = r"""
 ______  _  _   ____    ____   _____ 
|  ____|| || | / __ \  / __ \ |_   _|
| |__   | || || |  | || |  | |  | |  
|  __|  |__   || |  | || |  | |  | |  
| |____    | || |__| || |__| | _| |_ 
|______|   |_| \____/  \____/ |_____|
"""


from exaaiagnt.dashboard.server import start_dashboard


def should_use_live_display() -> bool:
    term = (os.getenv("TERM") or "").lower()
    if not sys.stdout.isatty():
        return False
    if term in {"", "dumb", "unknown"}:
        return False
    return True


async def run_cli(args: Any) -> None:  # noqa: PLR0915
    # Start the live dashboard
    try:
        start_dashboard()
        console_temp = Console()
        console_temp.print("[bold green]🚀 Live Dashboard available at: http://localhost:8000[/]")
    except Exception as e:
        import logging
        logging.exception(f"Failed to start dashboard: {e}")

    # Detect if running in a terminal that can reliably handle Live redraws
    is_tty = should_use_live_display()
    console = Console(force_terminal=is_tty, no_color=not is_tty)

    if is_tty:
        # Clear screen and show banner only in interactive terminal
        console.clear()
        console.print()
        console.print(BANNER, style="bold cyan", justify="center")
        console.print("[bold purple]Advanced AI-Powered Cybersecurity Agent[/]", justify="center")
        console.print("[dim]v2.2.7[/]", justify="center")
        console.print()
    else:
        # Simple text output for headless/pipe mode
        print("=" * 50)
        print("ExaAiAgent - AI-Powered Security Scanner")
        print("=" * 50)

    if is_tty:
        # Target info table (rich formatting)
        target_table = Table(show_header=True, header_style="bold cyan", border_style="cyan")
        target_table.add_column("Type", style="dim")
        target_table.add_column("Target", style="white")

        for target_info in args.targets_info:
            target_type = target_info.get("type", "URL")
            target_table.add_row(target_type, target_info["original"])

        console.print(Panel(target_table, title="[bold cyan]🎯 Targets", border_style="cyan"))
        console.print()

        # Config info
        config_text = Text()
        config_text.append("📁 Results: ", style="dim")
        config_text.append(f"exaai_runs/{args.run_name}\n", style="white")
        if args.instruction:
            config_text.append("📝 Instruction: ", style="dim")
            config_text.append(f"{args.instruction[:100]}{'...' if len(args.instruction) > 100 else ''}", style="white")

        console.print(Panel(config_text, title="[bold green]⚙️ Configuration", border_style="green"))
        console.print()
    else:
        # Simple text output for headless mode
        print(f"Targets: {[t['original'] for t in args.targets_info]}")
        print(f"Results: exaai_runs/{args.run_name}")
        if args.instruction:
            print(f"Instruction: {args.instruction[:100]}")
        print("-" * 50)

    scan_config = {
        "scan_id": args.run_name,
        "targets": args.targets_info,
        "user_instructions": args.instruction or "",
        "run_name": args.run_name,
    }

    # Handle prompt modules
    prompt_modules = None
    if getattr(args, "prompt_modules", None):
        prompt_modules = [m.strip() for m in args.prompt_modules.split(",")]

    llm_config = LLMConfig(prompt_modules=prompt_modules)
    agent_config = {
        "llm_config": llm_config,
        "max_iterations": 300,
        "non_interactive": True,
    }

    if getattr(args, "local_sources", None):
        agent_config["local_sources"] = args.local_sources

    tracer = Tracer(args.run_name)
    tracer.set_scan_config(scan_config)

    vuln_count = {"critical": 0, "high": 0, "medium": 0, "low": 0, "info": 0}

    def display_vulnerability(report_id: str, title: str, content: str, severity: str) -> None:
        severity_lower = severity.lower()
        vuln_count[severity_lower] = vuln_count.get(severity_lower, 0) + 1
        severity_color = get_severity_color(severity_lower)

        vuln_panel = Panel(
            Text.assemble(
                ("🔴 ", "bold red") if severity_lower in ["critical", "high"] else ("🟡 ", "bold yellow"),
                (title, "bold white"),
                ("\n\n", ""),
                (f"Severity: {severity.upper()}", f"bold {severity_color}"),
                ("\n\n", ""),
                (content[:500] + "..." if len(content) > 500 else content, "dim white"),
            ),
            title=f"[bold {severity_color}]🐞 {report_id.upper()}",
            border_style=severity_color,
            padding=(1, 2),
        )

        console.print(vuln_panel)
        console.print()

    tracer.vulnerability_found_callback = display_vulnerability

    def cleanup_on_exit() -> None:
        tracer.cleanup()

    def signal_handler(_signum: int, _frame: Any) -> None:
        console.print("\n[bold yellow]⚠️ Scan interrupted by user[/]")
        tracer.cleanup()
        sys.exit(1)

    atexit.register(cleanup_on_exit)
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    if hasattr(signal, "SIGHUP"):
        signal.signal(signal.SIGHUP, signal_handler)

    set_global_tracer(tracer)

    def create_live_status() -> Panel:
        status_text = Text()
        status_text.append("⏳ ", style="bold")
        status_text.append("Scanning in progress...", style="bold green")
        status_text.append("\n\n")

        # Live stats
        stats_text = build_live_stats_text(tracer, agent_config)
        if stats_text:
            status_text.append(stats_text)

        # Vulnerability summary
        total_vulns = sum(vuln_count.values())
        if total_vulns > 0:
            status_text.append("\n\n")
            status_text.append("Vulnerabilities Found: ", style="dim")
            status_text.append(f"{total_vulns}", style="bold red")

        return Panel(
            status_text,
            title="[bold green]🔍 Live Status",
            border_style="green",
            padding=(1, 2),
        )

    try:
        console.print("[bold cyan]🚀 Starting penetration test...[/]")
        console.print()

        if is_tty:
            with Live(
                create_live_status(),
                console=console,
                refresh_per_second=2,
                transient=False,
                auto_refresh=False,
            ) as live:
                stop_updates = threading.Event()

                def update_status() -> None:
                    while not stop_updates.is_set():
                        try:
                            live.update(create_live_status(), refresh=True)
                            time.sleep(2)
                        except Exception:  # noqa: BLE001
                            break

                update_thread = threading.Thread(target=update_status, daemon=True)
                update_thread.start()

                try:
                    agent = ExaaiAgent(agent_config)
                    result = await agent.execute_scan(scan_config)

                    if isinstance(result, dict) and not result.get("success", True):
                        error_msg = result.get("error", "Unknown error")
                        console.print()
                        console.print(f"[bold red]❌ Scan failed: {error_msg}[/]")
                        console.print()
                        sys.exit(1)
                finally:
                    stop_updates.set()
                    update_thread.join(timeout=1)
        else:
            agent = ExaaiAgent(agent_config)
            result = await agent.execute_scan(scan_config)

            if isinstance(result, dict) and not result.get("success", True):
                error_msg = result.get("error", "Unknown error")
                console.print()
                console.print(f"[bold red]❌ Scan failed: {error_msg}[/]")
                console.print()
                sys.exit(1)

    except Exception as e:
        tracer_error = ""
        if tracer and getattr(tracer, "agents", None):
            root_errors = [
                agent.get("error_message")
                for agent in tracer.agents.values()
                if agent.get("error_message")
            ]
            if root_errors:
                tracer_error = f"\n[dim]Agent error: {root_errors[-1]}[/]"

        console.print(f"[bold red]❌ Error: {e}[/]{tracer_error}")
        raise

    # Final Summary
    console.print()
    console.print("[bold green]" + "=" * 60 + "[/]")
    console.print("[bold green]✅ PENETRATION TEST COMPLETED[/]", justify="center")
    console.print("[bold green]" + "=" * 60 + "[/]")
    console.print()

    # Vulnerability Summary Table
    summary_table = Table(show_header=True, header_style="bold white", border_style="cyan")
    summary_table.add_column("Severity", style="bold")
    summary_table.add_column("Count", justify="center")

    summary_table.add_row("[red]Critical[/]", str(vuln_count.get("critical", 0)))
    summary_table.add_row("[orange1]High[/]", str(vuln_count.get("high", 0)))
    summary_table.add_row("[yellow]Medium[/]", str(vuln_count.get("medium", 0)))
    summary_table.add_row("[green]Low[/]", str(vuln_count.get("low", 0)))
    summary_table.add_row("[blue]Info[/]", str(vuln_count.get("info", 0)))

    total = sum(vuln_count.values())
    summary_table.add_row("[bold white]TOTAL[/]", f"[bold white]{total}[/]")

    console.print(Panel(summary_table, title="[bold cyan]📊 Vulnerability Summary", border_style="cyan"))

    # Final stats
    console.print()
    final_stats_text = build_final_stats_text(tracer)
    if final_stats_text:
        console.print(Panel(final_stats_text, title="[bold green]📈 Statistics", border_style="green"))

    # Final report
    if tracer.final_scan_result:
        console.print()
        console.print(Panel(
            tracer.final_scan_result,
            title="[bold cyan]📄 Final Report",
            border_style="cyan",
            padding=(1, 2),
        ))

    console.print()
    console.print(f"[dim]Results saved to: exaai_runs/{args.run_name}[/]")
    console.print()



import ipaddress
import re
import secrets
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import docker
from docker.errors import DockerException, ImageNotFound
from rich.console import Console
from rich.panel import Panel
from rich.text import Text


# Token formatting utilities
def format_token_count(count: float) -> str:
    count = int(count)
    if count >= 1_000_000:
        return f"{count / 1_000_000:.1f}M"
    if count >= 1_000:
        return f"{count / 1_000:.1f}K"
    return str(count)


# Display utilities
def get_severity_color(severity: str) -> str:
    severity_colors = {
        "critical": "#dc2626",
        "high": "#ea580c",
        "medium": "#d97706",
        "low": "#65a30d",
        "info": "#0284c7",
    }
    return severity_colors.get(severity, "#6b7280")


def _build_vulnerability_stats(stats_text: Text, tracer: Any) -> None:
    """Build vulnerability section of stats text."""
    vuln_count = len(tracer.vulnerability_reports)

    if vuln_count > 0:
        severity_counts = {"critical": 0, "high": 0, "medium": 0, "low": 0, "info": 0}
        for report in tracer.vulnerability_reports:
            severity = report.get("severity", "").lower()
            if severity in severity_counts:
                severity_counts[severity] += 1

        stats_text.append("🔍 Vulnerabilities Found: ", style="bold red")

        severity_parts = []
        for severity in ["critical", "high", "medium", "low", "info"]:
            count = severity_counts[severity]
            if count > 0:
                severity_color = get_severity_color(severity)
                severity_text = Text()
                severity_text.append(f"{severity.upper()}: ", style=severity_color)
                severity_text.append(str(count), style=f"bold {severity_color}")
                severity_parts.append(severity_text)

        for i, part in enumerate(severity_parts):
            stats_text.append(part)
            if i < len(severity_parts) - 1:
                stats_text.append(" | ", style="dim white")

        stats_text.append(" (Total: ", style="dim white")
        stats_text.append(str(vuln_count), style="bold yellow")
        stats_text.append(")", style="dim white")
        stats_text.append("\n")
    else:
        stats_text.append("🔍 Vulnerabilities Found: ", style="bold green")
        stats_text.append("0", style="bold white")
        stats_text.append(" (No exploitable vulnerabilities detected)", style="dim green")
        stats_text.append("\n")


def _build_llm_stats(stats_text: Text, total_stats: dict[str, Any]) -> None:
    """Build LLM usage section of stats text."""
    if total_stats["requests"] > 0:
        stats_text.append("\n")
        stats_text.append("📥 Input Tokens: ", style="bold cyan")
        stats_text.append(format_token_count(total_stats["input_tokens"]), style="bold white")

        if total_stats["cached_tokens"] > 0:
            stats_text.append(" • ", style="dim white")
            stats_text.append("⚡ Cached Tokens: ", style="bold green")
            stats_text.append(format_token_count(total_stats["cached_tokens"]), style="bold white")

        stats_text.append(" • ", style="dim white")
        stats_text.append("📤 Output Tokens: ", style="bold cyan")
        stats_text.append(format_token_count(total_stats["output_tokens"]), style="bold white")

        if total_stats["cost"] > 0:
            stats_text.append(" • ", style="dim white")
            stats_text.append("💰 Total Cost: ", style="bold cyan")
            stats_text.append(f"${total_stats['cost']:.4f}", style="bold yellow")
    else:
        stats_text.append("\n")
        stats_text.append("💰 Total Cost: ", style="bold cyan")
        stats_text.append("$0.0000 ", style="bold yellow")
        stats_text.append("• ", style="bold white")
        stats_text.append("📊 Tokens: ", style="bold cyan")
        stats_text.append("0", style="bold white")


def build_final_stats_text(tracer: Any) -> Text:
    """Build stats text for final output with detailed messages and LLM usage."""
    stats_text = Text()
    if not tracer:
        return stats_text

    _build_vulnerability_stats(stats_text, tracer)

    tool_count = tracer.get_real_tool_count()
    agent_count = len(tracer.agents)

    stats_text.append("🤖 Agents Used: ", style="bold cyan")
    stats_text.append(str(agent_count), style="bold white")
    stats_text.append(" • ", style="dim white")
    stats_text.append("🛠️ Tools Called: ", style="bold cyan")
    stats_text.append(str(tool_count), style="bold white")

    llm_stats = tracer.get_total_llm_stats()
    _build_llm_stats(stats_text, llm_stats["total"])

    return stats_text


def build_live_stats_text(tracer: Any, agent_config: dict[str, Any] | None = None) -> Text:
    stats_text = Text()
    if not tracer:
        return stats_text

    vuln_count = len(tracer.vulnerability_reports)
    tool_count = tracer.get_real_tool_count()
    agent_count = len(tracer.agents)

    stats_text.append("🔍 Vulnerabilities: ", style="bold white")
    stats_text.append(f"{vuln_count}", style="dim white")
    stats_text.append("\n")
    if vuln_count > 0:
        severity_counts = {"critical": 0, "high": 0, "medium": 0, "low": 0, "info": 0}
        for report in tracer.vulnerability_reports:
            severity = report.get("severity", "").lower()
            if severity in severity_counts:
                severity_counts[severity] += 1

        severity_parts = []
        for severity in ["critical", "high", "medium", "low", "info"]:
            count = severity_counts[severity]
            if count > 0:
                severity_color = get_severity_color(severity)
                severity_text = Text()
                severity_text.append(f"{severity.upper()}: ", style=severity_color)
                severity_text.append(str(count), style=f"bold {severity_color}")
                severity_parts.append(severity_text)

        for i, part in enumerate(severity_parts):
            stats_text.append(part)
            if i < len(severity_parts) - 1:
                stats_text.append(" | ", style="dim white")

        stats_text.append("\n")

    if agent_config:
        llm_config = agent_config["llm_config"]
        model = getattr(llm_config, "model_name", "Unknown")
        stats_text.append(f"🧠 Model: {model}")
        stats_text.append("\n")

    stats_text.append("🤖 Agents: ", style="bold white")
    stats_text.append(str(agent_count), style="dim white")
    stats_text.append(" • ", style="dim white")
    stats_text.append("🛠️ Tools: ", style="bold white")
    stats_text.append(str(tool_count), style="dim white")

    llm_stats = tracer.get_total_llm_stats()
    total_stats = llm_stats["total"]

    stats_text.append("\n")

    stats_text.append("📥 Input: ", style="bold white")
    stats_text.append(format_token_count(total_stats["input_tokens"]), style="dim white")

    stats_text.append(" • ", style="dim white")
    stats_text.append("⚡ ", style="bold white")
    stats_text.append("Cached: ", style="bold white")
    stats_text.append(format_token_count(total_stats["cached_tokens"]), style="dim white")

    stats_text.append("\n")

    stats_text.append("📤 Output: ", style="bold white")
    stats_text.append(format_token_count(total_stats["output_tokens"]), style="dim white")

    stats_text.append(" • ", style="dim white")
    stats_text.append("💰 Cost: ", style="bold white")
    stats_text.append(f"${total_stats['cost']:.4f}", style="dim white")

    return stats_text


# Name generation utilities


def _slugify_for_run_name(text: str, max_length: int = 32) -> str:
    text = text.lower().strip()
    text = re.sub(r"[^a-z0-9]+", "-", text)
    text = text.strip("-")
    if len(text) > max_length:
        text = text[:max_length].rstrip("-")
    return text or "pentest"


def _derive_target_label_for_run_name(targets_info: list[dict[str, Any]] | None) -> str:  # noqa: PLR0911
    if not targets_info:
        return "pentest"

    first = targets_info[0]
    target_type = first.get("type")
    details = first.get("details", {}) or {}
    original = first.get("original", "") or ""

    if target_type == "web_application":
        url = details.get("target_url", original)
        try:
            parsed = urlparse(url)
            return str(parsed.netloc or parsed.path or url)
        except Exception:  # noqa: BLE001
            return str(url)

    if target_type == "repository":
        repo = details.get("target_repo", original)
        parsed = urlparse(repo)
        path = parsed.path or repo
        name = path.rstrip("/").split("/")[-1] or path
        if name.endswith(".git"):
            name = name[:-4]
        return str(name)

    if target_type == "local_code":
        path_str = details.get("target_path", original)
        try:
            return str(Path(path_str).name or path_str)
        except Exception:  # noqa: BLE001
            return str(path_str)

    if target_type == "ip_address":
        return str(details.get("target_ip", original) or original)

    return str(original or "pentest")


def generate_run_name(targets_info: list[dict[str, Any]] | None = None) -> str:
    base_label = _derive_target_label_for_run_name(targets_info)
    slug = _slugify_for_run_name(base_label)

    random_suffix = secrets.token_hex(2)

    return f"{slug}_{random_suffix}"


# Target processing utilities
def infer_target_type(target: str) -> tuple[str, dict[str, str]]:  # noqa: PLR0911
    if not target or not isinstance(target, str):
        raise ValueError("Target must be a non-empty string")

    target = target.strip()

    lower_target = target.lower()
    bare_repo_prefixes = (
        "github.com/",
        "www.github.com/",
        "gitlab.com/",
        "www.gitlab.com/",
        "bitbucket.org/",
        "www.bitbucket.org/",
    )
    if any(lower_target.startswith(p) for p in bare_repo_prefixes):
        return "repository", {"target_repo": f"https://{target}"}

    parsed = urlparse(target)
    if parsed.scheme in ("http", "https"):
        if any(
            host in parsed.netloc.lower() for host in ["github.com", "gitlab.com", "bitbucket.org"]
        ):
            return "repository", {"target_repo": target}
        return "web_application", {"target_url": target}

    try:
        ip_obj = ipaddress.ip_address(target)
    except ValueError:
        pass
    else:
        return "ip_address", {"target_ip": str(ip_obj)}

    path = Path(target).expanduser()
    try:
        if path.exists():
            if path.is_dir():
                resolved = path.resolve()
                return "local_code", {"target_path": str(resolved)}
            raise ValueError(f"Path exists but is not a directory: {target}")
    except (OSError, RuntimeError) as e:
        raise ValueError(f"Invalid path: {target} - {e!s}") from e

    if target.startswith("git@") or target.endswith(".git"):
        return "repository", {"target_repo": target}

    if "." in target and "/" not in target and not target.startswith("."):
        parts = target.split(".")
        if len(parts) >= 2 and all(p and p.strip() for p in parts):
            return "web_application", {"target_url": f"https://{target}"}

    raise ValueError(
        f"Invalid target: {target}\n"
        "Target must be one of:\n"
        "- A valid URL (http:// or https://)\n"
        "- A Git repository URL (https://github.com/... or git@github.com:...)\n"
        "- A local directory path\n"
        "- A domain name (e.g., example.com)\n"
        "- An IP address (e.g., 192.168.1.10)"
    )


def sanitize_name(name: str) -> str:
    sanitized = re.sub(r"[^A-Za-z0-9._-]", "-", name.strip())
    return sanitized or "target"


def derive_repo_base_name(repo_url: str) -> str:
    if repo_url.endswith("/"):
        repo_url = repo_url[:-1]

    if ":" in repo_url and repo_url.startswith("git@"):
        path_part = repo_url.split(":", 1)[1]
    else:
        path_part = urlparse(repo_url).path or repo_url

    candidate = path_part.split("/")[-1]
    if candidate.endswith(".git"):
        candidate = candidate[:-4]

    return sanitize_name(candidate or "repository")


def derive_local_base_name(path_str: str) -> str:
    try:
        base = Path(path_str).resolve().name
    except (OSError, RuntimeError):
        base = Path(path_str).name
    return sanitize_name(base or "workspace")


def assign_workspace_subdirs(targets_info: list[dict[str, Any]]) -> None:
    name_counts: dict[str, int] = {}

    for target in targets_info:
        target_type = target["type"]
        details = target["details"]

        base_name: str | None = None
        if target_type == "repository":
            base_name = derive_repo_base_name(details["target_repo"])
        elif target_type == "local_code":
            base_name = derive_local_base_name(details.get("target_path", "local"))

        if base_name is None:
            continue

        count = name_counts.get(base_name, 0) + 1
        name_counts[base_name] = count

        workspace_subdir = base_name if count == 1 else f"{base_name}-{count}"

        details["workspace_subdir"] = workspace_subdir


def collect_local_sources(targets_info: list[dict[str, Any]]) -> list[dict[str, str]]:
    local_sources: list[dict[str, str]] = []

    for target_info in targets_info:
        details = target_info["details"]
        workspace_subdir = details.get("workspace_subdir")

        if target_info["type"] == "local_code" and "target_path" in details:
            local_sources.append(
                {
                    "source_path": details["target_path"],
                    "workspace_subdir": workspace_subdir,
                }
            )

        elif target_info["type"] == "repository" and "cloned_repo_path" in details:
            local_sources.append(
                {
                    "source_path": details["cloned_repo_path"],
                    "workspace_subdir": workspace_subdir,
                }
            )

    return local_sources


# Repository utilities
def clone_repository(repo_url: str, run_name: str, dest_name: str | None = None) -> str:
    console = Console()

    git_executable = shutil.which("git")
    if git_executable is None:
        raise FileNotFoundError("Git executable not found in PATH")

    temp_dir = Path(tempfile.gettempdir()) / "exaai_repos" / run_name
    temp_dir.mkdir(parents=True, exist_ok=True)

    if dest_name:
        repo_name = dest_name
    else:
        repo_name = Path(repo_url).stem if repo_url.endswith(".git") else Path(repo_url).name

    clone_path = temp_dir / repo_name

    if clone_path.exists():
        shutil.rmtree(clone_path)

    try:
        with console.status(f"[bold cyan]Cloning repository {repo_url}...", spinner="dots"):
            subprocess.run(  # noqa: S603
                [
                    git_executable,
                    "clone",
                    repo_url,
                    str(clone_path),
                ],
                capture_output=True,
                text=True,
                check=True,
            )

        return str(clone_path.absolute())

    except subprocess.CalledProcessError as e:
        error_text = Text()
        error_text.append("❌ ", style="bold red")
        error_text.append("REPOSITORY CLONE FAILED", style="bold red")
        error_text.append("\n\n", style="white")
        error_text.append(f"Could not clone repository: {repo_url}\n", style="white")
        error_text.append(
            f"Error: {e.stderr if hasattr(e, 'stderr') and e.stderr else str(e)}", style="dim red"
        )

        panel = Panel(
            error_text,
            title="[bold red]🛡️  EXAAI CLONE ERROR",
            title_align="center",
            border_style="red",
            padding=(1, 2),
        )
        console.print("\n")
        console.print(panel)
        console.print()
        sys.exit(1)
    except FileNotFoundError:
        error_text = Text()
        error_text.append("❌ ", style="bold red")
        error_text.append("GIT NOT FOUND", style="bold red")
        error_text.append("\n\n", style="white")
        error_text.append("Git is not installed or not available in PATH.\n", style="white")
        error_text.append("Please install Git to clone repositories.\n", style="white")

        panel = Panel(
            error_text,
            title="[bold red]🛡️  EXAAI CLONE ERROR",
            title_align="center",
            border_style="red",
            padding=(1, 2),
        )
        console.print("\n")
        console.print(panel)
        console.print()
        sys.exit(1)


# Docker utilities
def check_docker_connection() -> Any:
    try:
        return docker.from_env()
    except DockerException:
        console = Console()
        error_text = Text()
        error_text.append("❌ ", style="bold red")
        error_text.append("DOCKER NOT AVAILABLE", style="bold red")
        error_text.append("\n\n", style="white")
        error_text.append("Cannot connect to Docker daemon.\n", style="white")
        error_text.append("Please ensure Docker is installed and running.\n\n", style="white")
        error_text.append("Try running: ", style="dim white")
        error_text.append("sudo systemctl start docker", style="dim cyan")

        panel = Panel(
            error_text,
            title="[bold red]🛡️  EXAAI STARTUP ERROR",
            title_align="center",
            border_style="red",
            padding=(1, 2),
        )
        console.print("\n", panel, "\n")
        raise RuntimeError("Docker not available") from None


def image_exists(client: Any, image_name: str) -> bool:
    try:
        client.images.get(image_name)
    except ImageNotFound:
        return False
    else:
        return True


def update_layer_status(layers_info: dict[str, str], layer_id: str, layer_status: str) -> None:
    if "Pull complete" in layer_status or "Already exists" in layer_status:
        layers_info[layer_id] = "✓"
    elif "Downloading" in layer_status:
        layers_info[layer_id] = "↓"
    elif "Extracting" in layer_status:
        layers_info[layer_id] = "📦"
    elif "Waiting" in layer_status:
        layers_info[layer_id] = "⏳"
    else:
        layers_info[layer_id] = "•"


def process_pull_line(
    line: dict[str, Any], layers_info: dict[str, str], status: Any, last_update: str
) -> str:
    if "id" in line and "status" in line:
        layer_id = line["id"]
        update_layer_status(layers_info, layer_id, line["status"])

        completed = sum(1 for v in layers_info.values() if v == "✓")
        total = len(layers_info)

        if total > 0:
            update_msg = f"[bold cyan]Progress: {completed}/{total} layers complete"
            if update_msg != last_update:
                status.update(update_msg)
                return update_msg

    elif "status" in line and "id" not in line:
        global_status = line["status"]
        if "Pulling from" in global_status:
            status.update("[bold cyan]Fetching image manifest...")
        elif "Digest:" in global_status:
            status.update("[bold cyan]Verifying image...")
        elif "Status:" in global_status:
            status.update("[bold cyan]Finalizing...")

    return last_update


# LLM utilities
def validate_llm_response(response: Any) -> None:
    if not response or not response.choices or not response.choices[0].message.content:
        raise RuntimeError("Invalid response from LLM")



import atexit
import contextlib
import signal
import sys
import threading
from typing import Any

from .browser_instance import BrowserInstance


class BrowserTabManager:
    def __init__(self) -> None:
        self.browser_instance: BrowserInstance | None = None
        self._lock = threading.Lock()

        self._register_cleanup_handlers()

    def launch_browser(self, url: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is not None:
                raise ValueError("Browser is already launched")

            try:
                self.browser_instance = BrowserInstance()
                result = self.browser_instance.launch(url)
                result["message"] = "Browser launched successfully"
            except (OSError, ValueError, RuntimeError) as e:
                if self.browser_instance:
                    self.browser_instance = None
                raise RuntimeError(f"Failed to launch browser: {e}") from e
            else:
                return result

    def goto_url(self, url: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.goto(url, tab_id)
            result["message"] = f"Navigated to {url}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to navigate to URL: {e}") from e
        else:
            return result

    def click(self, coordinate: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.click(coordinate, tab_id)
            result["message"] = f"Clicked at {coordinate}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to click: {e}") from e
        else:
            return result

    def type_text(self, text: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.type_text(text, tab_id)
            result["message"] = f"Typed text: {text[:50]}{'...' if len(text) > 50 else ''}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to type text: {e}") from e
        else:
            return result

    def scroll(self, direction: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.scroll(direction, tab_id)
            result["message"] = f"Scrolled {direction}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to scroll: {e}") from e
        else:
            return result

    def back(self, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.back(tab_id)
            result["message"] = "Navigated back"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to go back: {e}") from e
        else:
            return result

    def forward(self, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.forward(tab_id)
            result["message"] = "Navigated forward"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to go forward: {e}") from e
        else:
            return result

    def new_tab(self, url: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.new_tab(url)
            result["message"] = f"Created new tab {result.get('tab_id', '')}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to create new tab: {e}") from e
        else:
            return result

    def switch_tab(self, tab_id: str) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.switch_tab(tab_id)
            result["message"] = f"Switched to tab {tab_id}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to switch tab: {e}") from e
        else:
            return result

    def close_tab(self, tab_id: str) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.close_tab(tab_id)
            result["message"] = f"Closed tab {tab_id}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to close tab: {e}") from e
        else:
            return result

    def wait_browser(self, duration: float, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.wait(duration, tab_id)
            result["message"] = f"Waited {duration}s"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to wait: {e}") from e
        else:
            return result

    def execute_js(self, js_code: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.execute_js(js_code, tab_id)
            result["message"] = "JavaScript executed successfully"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to execute JavaScript: {e}") from e
        else:
            return result

    def double_click(self, coordinate: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.double_click(coordinate, tab_id)
            result["message"] = f"Double clicked at {coordinate}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to double click: {e}") from e
        else:
            return result

    def hover(self, coordinate: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.hover(coordinate, tab_id)
            result["message"] = f"Hovered at {coordinate}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to hover: {e}") from e
        else:
            return result

    def press_key(self, key: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.press_key(key, tab_id)
            result["message"] = f"Pressed key {key}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to press key: {e}") from e
        else:
            return result

    def save_pdf(self, file_path: str, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.save_pdf(file_path, tab_id)
            result["message"] = f"Page saved as PDF: {file_path}"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to save PDF: {e}") from e
        else:
            return result

    def get_console_logs(self, tab_id: str | None = None, clear: bool = False) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.get_console_logs(tab_id, clear)
            action_text = "cleared and retrieved" if clear else "retrieved"

            logs = result.get("console_logs", [])
            truncated = any(log.get("text", "").startswith("[TRUNCATED:") for log in logs)
            truncated_text = " (truncated)" if truncated else ""

            result["message"] = (
                f"Console logs {action_text} for tab "
                f"{result.get('tab_id', 'current')}{truncated_text}"
            )
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to get console logs: {e}") from e
        else:
            return result

    def view_source(self, tab_id: str | None = None) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

        try:
            result = self.browser_instance.view_source(tab_id)
            result["message"] = "Page source retrieved"
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to get page source: {e}") from e
        else:
            return result

    def list_tabs(self) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                return {"tabs": {}, "total_count": 0, "current_tab": None}

        try:
            tab_info = {}
            for tid, tab_page in self.browser_instance.pages.items():
                try:
                    tab_info[tid] = {
                        "url": tab_page.url,
                        "title": "Unknown" if tab_page.is_closed() else "Active",
                        "is_current": tid == self.browser_instance.current_page_id,
                    }
                except (AttributeError, RuntimeError):
                    tab_info[tid] = {
                        "url": "Unknown",
                        "title": "Closed",
                        "is_current": False,
                    }

            return {
                "tabs": tab_info,
                "total_count": len(tab_info),
                "current_tab": self.browser_instance.current_page_id,
            }
        except (OSError, ValueError, RuntimeError) as e:
            raise RuntimeError(f"Failed to list tabs: {e}") from e

    def close_browser(self) -> dict[str, Any]:
        with self._lock:
            if self.browser_instance is None:
                raise ValueError("Browser not launched")

            try:
                self.browser_instance.close()
                self.browser_instance = None
            except (OSError, ValueError, RuntimeError) as e:
                raise RuntimeError(f"Failed to close browser: {e}") from e
            else:
                return {
                    "message": "Browser closed successfully",
                    "screenshot": "",
                    "is_running": False,
                }

    def cleanup_dead_browser(self) -> None:
        with self._lock:
            if self.browser_instance and not self.browser_instance.is_alive():
                with contextlib.suppress(Exception):
                    self.browser_instance.close()
                self.browser_instance = None

    def close_all(self) -> None:
        with self._lock:
            if self.browser_instance:
                with contextlib.suppress(Exception):
                    self.browser_instance.close()
                self.browser_instance = None

    def _register_cleanup_handlers(self) -> None:
        atexit.register(self.close_all)

        signal.signal(signal.SIGTERM, self._signal_handler)
        signal.signal(signal.SIGINT, self._signal_handler)

        if hasattr(signal, "SIGHUP"):
            signal.signal(signal.SIGHUP, self._signal_handler)

    def _signal_handler(self, _signum: int, _frame: Any) -> None:
        self.close_all()
        sys.exit(0)


_browser_tab_manager = BrowserTabManager()


def get_browser_tab_manager() -> BrowserTabManager:
    return _browser_tab_manager



"""
Enhanced LLM Request Queue with intelligent rate limiting, retry logic, and fallback support.
"""

import asyncio
import logging
import os
import random
import threading
import time
from typing import Any

import litellm
from litellm import ModelResponse, completion
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential,
)


logger = logging.getLogger(__name__)


def should_retry_exception(exception: Exception) -> bool:
    """Determine if an exception should trigger a retry."""
    status_code = None

    if hasattr(exception, "status_code"):
        status_code = exception.status_code
    elif hasattr(exception, "response") and hasattr(exception.response, "status_code"):
        status_code = exception.response.status_code

    # Always retry rate limit errors
    if isinstance(exception, litellm.RateLimitError):
        logger.warning("Rate limit hit, will retry with backoff...")
        return True

    # Retry on timeout
    if isinstance(exception, (litellm.Timeout, asyncio.TimeoutError)):
        logger.warning("Request timeout, will retry...")
        return True

    # Retry on server errors
    if isinstance(exception, (litellm.ServiceUnavailableError, litellm.InternalServerError)):
        logger.warning("Server error, will retry...")
        return True

    if status_code is not None:
        return bool(litellm._should_retry(status_code))

    return True


class LLMRequestQueue:
    """
    Enhanced request queue with:
    - Intelligent rate limiting with jitter
    - Exponential backoff on failures
    - Fallback model support
    - Request timeout handling
    """

    def __init__(
        self,
        max_concurrent: int = 2,
        delay_between_requests: float = 4.0,
        request_timeout: int = 300,
    ):
        # Load configuration from environment
        rate_limit_delay = os.getenv("LLM_RATE_LIMIT_DELAY")
        if rate_limit_delay:
            delay_between_requests = float(rate_limit_delay)

        rate_limit_concurrent = os.getenv("LLM_RATE_LIMIT_CONCURRENT")
        if rate_limit_concurrent:
            max_concurrent = int(rate_limit_concurrent)

        timeout_env = os.getenv("LLM_REQUEST_TIMEOUT")
        if timeout_env:
            request_timeout = int(timeout_env)

        self.max_concurrent = max_concurrent
        self.delay_between_requests = delay_between_requests
        self.request_timeout = request_timeout
        self._semaphore = threading.BoundedSemaphore(max_concurrent)
        self._last_request_time = 0.0
        self._lock = threading.Lock()
        self._consecutive_failures = 0
        self._max_consecutive_failures = 5

        # Fallback models configuration
        self._fallback_models = self._load_fallback_models()

        logger.info(
            f"LLM Queue initialized: concurrent={max_concurrent}, "
            f"delay={delay_between_requests}s, timeout={request_timeout}s"
        )

    def _load_fallback_models(self) -> list[str]:
        """Load fallback models from environment."""
        fallback_env = os.getenv("EXAAI_FALLBACK_MODELS", "")
        if fallback_env:
            return [m.strip() for m in fallback_env.split(",") if m.strip()]
        return []

    def _add_jitter(self, delay: float) -> float:
        """Add random jitter to delay to prevent thundering herd."""
        jitter = random.uniform(0.1, 0.5) * delay
        return delay + jitter

    async def make_request(self, completion_args: dict[str, Any]) -> ModelResponse:
        """Make an LLM request with rate limiting and retry logic."""
        try:
            # Acquire semaphore with timeout
            acquire_start = time.time()
            while not self._semaphore.acquire(timeout=0.5):
                if time.time() - acquire_start > 60:
                    raise TimeoutError("Failed to acquire request slot after 60s")
                await asyncio.sleep(0.2)

            # Calculate delay with adaptive backoff
            with self._lock:
                now = time.time()
                time_since_last = now - self._last_request_time

                # Increase delay if we've had failures
                adaptive_delay = self.delay_between_requests
                if self._consecutive_failures > 0:
                    adaptive_delay *= (1.5 ** self._consecutive_failures)
                    adaptive_delay = min(adaptive_delay, 30.0)  # Cap at 30s

                sleep_needed = max(0, adaptive_delay - time_since_last)
                sleep_needed = self._add_jitter(sleep_needed)
                self._last_request_time = now + sleep_needed

            if sleep_needed > 0:
                logger.debug(f"Rate limiting: waiting {sleep_needed:.2f}s")
                await asyncio.sleep(sleep_needed)

            # Make the request with timeout
            try:
                response = await asyncio.wait_for(
                    self._reliable_request(completion_args),
                    timeout=self.request_timeout
                )
                # Reset failure counter on success
                with self._lock:
                    self._consecutive_failures = 0
                return response

            except TimeoutError:
                logger.error(f"Request timed out after {self.request_timeout}s")
                with self._lock:
                    self._consecutive_failures += 1
                raise litellm.Timeout(f"Request timed out after {self.request_timeout}s")

        finally:
            self._semaphore.release()

    @retry(
        stop=stop_after_attempt(15),  # Increased for rate limits
        wait=wait_exponential(multiplier=3, min=10, max=180),  # Longer delays
        retry=retry_if_exception(should_retry_exception),
        before_sleep=before_sleep_log(logger, logging.WARNING),
        reraise=True,
    )
    async def _reliable_request(self, completion_args: dict[str, Any]) -> ModelResponse:
        """Make a reliable request with automatic retry."""
        try:
            response = completion(**completion_args, stream=False)
            if isinstance(response, ModelResponse):
                return response
            self._raise_unexpected_response()
            raise RuntimeError("Unreachable code")

        except litellm.RateLimitError:
            # Track consecutive failures
            with self._lock:
                self._consecutive_failures += 1

            # Try fallback model if available and we've failed multiple times
            if self._consecutive_failures >= 3 and self._fallback_models:
                fallback_response = await self._try_fallback_models(completion_args)
                if fallback_response:
                    return fallback_response
            raise

        except (litellm.ServiceUnavailableError, litellm.InternalServerError):
            with self._lock:
                self._consecutive_failures += 1
            raise

    async def _try_fallback_models(
        self, completion_args: dict[str, Any]
    ) -> ModelResponse | None:
        """Try fallback models when primary fails."""
        original_model = completion_args.get("model", "")

        for fallback_model in self._fallback_models:
            if fallback_model == original_model:
                continue

            logger.warning(f"Trying fallback model: {fallback_model}")
            try:
                fallback_args = completion_args.copy()
                fallback_args["model"] = fallback_model

                response = completion(**fallback_args, stream=False)
                if isinstance(response, ModelResponse):
                    logger.info(f"Fallback model {fallback_model} succeeded")
                    return response

            except Exception as e:
                logger.warning(f"Fallback model {fallback_model} failed: {e}")
                continue

        return None

    def _raise_unexpected_response(self) -> None:
        raise RuntimeError("Unexpected response type")

    def get_stats(self) -> dict[str, Any]:
        """Get queue statistics."""
        return {
            "consecutive_failures": self._consecutive_failures,
            "max_concurrent": self.max_concurrent,
            "delay_between_requests": self.delay_between_requests,
            "fallback_models": self._fallback_models,
        }


_global_queue: LLMRequestQueue | None = None


def get_global_queue() -> LLMRequestQueue:
    global _global_queue
    if _global_queue is None:
        _global_queue = LLMRequestQueue()
    return _global_queue


def reset_global_queue() -> None:
    """Reset the global queue (useful for testing or reconfiguration)."""
    global _global_queue
    _global_queue = None



import json
import re
from pathlib import Path
from typing import Any, cast

from openhands_aci import file_editor
from openhands_aci.utils.shell import run_shell_cmd

from exaaiagnt.tools.registry import register_tool


def _parse_file_editor_output(output: str) -> dict[str, Any]:
    try:
        pattern = r"<oh_aci_output_[^>]+>\n(.*?)\n</oh_aci_output_[^>]+>"
        match = re.search(pattern, output, re.DOTALL)

        if match:
            json_str = match.group(1)
            data = json.loads(json_str)
            return cast("dict[str, Any]", data)
        return {"output": output, "error": None}
    except (json.JSONDecodeError, AttributeError):
        return {"output": output, "error": None}


@register_tool
def str_replace_editor(
    command: str,
    path: str,
    file_text: str | None = None,
    view_range: list[int] | None = None,
    old_str: str | None = None,
    new_str: str | None = None,
    insert_line: int | None = None,
) -> dict[str, Any]:
    try:
        path_obj = Path(path)
        if not path_obj.is_absolute():
            path = str(Path("/workspace") / path_obj)

        result = file_editor(
            command=command,
            path=path,
            file_text=file_text,
            view_range=view_range,
            old_str=old_str,
            new_str=new_str,
            insert_line=insert_line,
        )

        parsed = _parse_file_editor_output(result)

        if parsed.get("error"):
            return {"error": parsed["error"]}

        return {"content": parsed.get("output", result)}

    except (OSError, ValueError) as e:
        return {"error": f"Error in {command} operation: {e!s}"}


@register_tool
def list_files(
    path: str,
    recursive: bool = False,
) -> dict[str, Any]:
    try:
        path_obj = Path(path)
        if not path_obj.is_absolute():
            path = str(Path("/workspace") / path_obj)
            path_obj = Path(path)

        if not path_obj.exists():
            return {"error": f"Directory not found: {path}"}

        if not path_obj.is_dir():
            return {"error": f"Path is not a directory: {path}"}

        cmd = f"find '{path}' -type f -o -type d | head -500" if recursive else f"ls -1a '{path}'"

        exit_code, stdout, stderr = run_shell_cmd(cmd)

        if exit_code != 0:
            return {"error": f"Error listing directory: {stderr}"}

        items = stdout.strip().split("\n") if stdout.strip() else []

        files = []
        dirs = []

        for item in items:
            item_path = item if recursive else str(Path(path) / item)
            item_path_obj = Path(item_path)

            if item_path_obj.is_file():
                files.append(item)
            elif item_path_obj.is_dir():
                dirs.append(item)

        return {
            "files": sorted(files),
            "directories": sorted(dirs),
            "total_files": len(files),
            "total_dirs": len(dirs),
            "path": path,
            "recursive": recursive,
        }

    except (OSError, ValueError) as e:
        return {"error": f"Error listing directory: {e!s}"}


@register_tool
def search_files(
    path: str,
    regex: str,
    file_pattern: str = "*",
) -> dict[str, Any]:
    try:
        path_obj = Path(path)
        if not path_obj.is_absolute():
            path = str(Path("/workspace") / path_obj)

        if not Path(path).exists():
            return {"error": f"Directory not found: {path}"}

        escaped_regex = regex.replace("'", "'\"'\"'")

        cmd = f"rg --line-number --glob '{file_pattern}' '{escaped_regex}' '{path}'"

        exit_code, stdout, stderr = run_shell_cmd(cmd)

        if exit_code not in {0, 1}:
            return {"error": f"Error searching files: {stderr}"}
        return {"output": stdout if stdout else "No matches found"}

    except (OSError, ValueError) as e:
        return {"error": f"Error searching files: {e!s}"}


# ruff: noqa: TRY300



"""
WAF Detection & Bypass Module - Detects and bypasses Web Application Firewalls.
Supports Cloudflare, Akamai, Imperva, AWS WAF, and more.
"""

import base64
import json
import logging
import re
import urllib.parse
from dataclasses import dataclass
from enum import Enum


logger = logging.getLogger(__name__)


class WAFType(Enum):
    """Known WAF types."""
    CLOUDFLARE = "cloudflare"
    AKAMAI = "akamai"
    IMPERVA = "imperva"
    AWS_WAF = "aws_waf"
    MODSECURITY = "modsecurity"
    F5_BIG_IP = "f5_big_ip"
    SUCURI = "sucuri"
    WORDFENCE = "wordfence"
    BARRACUDA = "barracuda"
    FORTIWEB = "fortiweb"
    UNKNOWN = "unknown"


@dataclass
class WAFDetectionResult:
    """Result of WAF detection."""
    detected: bool
    waf_type: WAFType
    confidence: float
    indicators: list[str]
    bypass_suggestions: list[str]


class WAFDetector:
    """Detects WAF presence based on response characteristics."""

    # Detection patterns
    DETECTION_PATTERNS: dict[WAFType, dict[str, list[str]]] = {
        WAFType.CLOUDFLARE: {
            "headers": ["cf-ray", "cf-cache-status", "__cfduid", "cf-request-id"],
            "cookies": ["__cfduid", "__cf_bm", "cf_clearance"],
            "body": ["cloudflare", "cf-browser-verification", "attention required"],
            "status_pages": ["error 1020", "ray id"],
        },
        WAFType.AKAMAI: {
            "headers": ["akamai", "x-akamai", "akamai-grn"],
            "cookies": ["akamai", "_abck", "bm_sz", "ak_bmsc"],
            "body": ["akamai", "access denied", "reference#"],
        },
        WAFType.IMPERVA: {
            "headers": ["x-iinfo", "x-cdn"],
            "cookies": ["incap_ses", "visid_incap", "nlbi_"],
            "body": ["incapsula", "imperva", "request unsuccessful"],
        },
        WAFType.AWS_WAF: {
            "headers": ["x-amzn-requestid", "x-amz-cf-id"],
            "body": ["aws", "request blocked", "waf"],
        },
        WAFType.MODSECURITY: {
            "headers": ["mod_security", "modsec"],
            "body": ["modsecurity", "not acceptable", "mod_security"],
        },
        WAFType.SUCURI: {
            "headers": ["x-sucuri-id", "x-sucuri-cache"],
            "body": ["sucuri", "cloudproxy"],
        },
    }

    def detect(
        self,
        status_code: int,
        headers: dict[str, str],
        body: str,
        cookies: dict[str, str] = None
    ) -> WAFDetectionResult:
        """Detect WAF from response."""
        cookies = cookies or {}
        indicators = []
        detected_waf = WAFType.UNKNOWN
        max_confidence = 0.0

        # Check for blocking status codes
        if status_code in [403, 406, 429, 503]:
            indicators.append(f"Blocking status code: {status_code}")

        # Check each WAF type
        for waf_type, patterns in self.DETECTION_PATTERNS.items():
            confidence = 0.0
            waf_indicators = []

            # Check headers
            headers_lower = {k.lower(): v.lower() for k, v in headers.items()}
            for pattern in patterns.get("headers", []):
                if any(pattern in h for h in headers_lower):
                    confidence += 0.3
                    waf_indicators.append(f"Header match: {pattern}")

            # Check cookies
            cookies_lower = {k.lower(): v.lower() for k, v in cookies.items()}
            for pattern in patterns.get("cookies", []):
                if any(pattern in c for c in cookies_lower):
                    confidence += 0.25
                    waf_indicators.append(f"Cookie match: {pattern}")

            # Check body
            body_lower = body.lower()
            for pattern in patterns.get("body", []):
                if pattern in body_lower:
                    confidence += 0.2
                    waf_indicators.append(f"Body match: {pattern}")

            if confidence > max_confidence:
                max_confidence = confidence
                detected_waf = waf_type
                indicators = waf_indicators

        detected = max_confidence > 0.3 or status_code in [403, 406]

        return WAFDetectionResult(
            detected=detected,
            waf_type=detected_waf,
            confidence=min(max_confidence, 1.0),
            indicators=indicators,
            bypass_suggestions=self._get_bypass_suggestions(detected_waf) if detected else []
        )

    def _get_bypass_suggestions(self, waf_type: WAFType) -> list[str]:
        """Get bypass suggestions for detected WAF."""
        common = [
            "Try Unicode normalization",
            "Use alternate encodings (double URL, hex)",
            "Try HTTP parameter pollution",
            "Use case variation",
            "Add comment injection",
        ]

        specific = {
            WAFType.CLOUDFLARE: [
                "Use IP-based bypass if origin exposed",
                "Try header smuggling",
                "Use chunked encoding",
            ],
            WAFType.AKAMAI: [
                "Try JSON smuggling",
                "Use multipart encoding",
                "Header padding techniques",
            ],
            WAFType.IMPERVA: [
                "Try wildcard techniques",
                "Use nested encoding",
            ],
        }

        return common + specific.get(waf_type, [])


class PayloadEncoder:
    """Encodes payloads to bypass WAF."""

    @staticmethod
    def url_encode(payload: str, double: bool = False) -> str:
        """URL encode payload."""
        encoded = urllib.parse.quote(payload, safe="")
        if double:
            encoded = urllib.parse.quote(encoded, safe="")
        return encoded

    @staticmethod
    def unicode_normalize(payload: str) -> str:
        """Apply Unicode normalization."""
        replacements = {
            "<": "\uff1c",
            ">": "\uff1e",
            "'": "\u2019",
            '"': "\u201c",
            "(": "\uff08",
            ")": "\uff09",
            "/": "\u2215",
        }
        result = payload
        for char, replacement in replacements.items():
            result = result.replace(char, replacement)
        return result

    @staticmethod
    def hex_encode(payload: str) -> str:
        """Hex encode payload."""
        return "".join(f"%{ord(c):02x}" for c in payload)

    @staticmethod
    def base64_encode(payload: str) -> str:
        """Base64 encode payload."""
        return base64.b64encode(payload.encode()).decode()

    @staticmethod
    def add_comments(payload: str, style: str = "sql") -> str:
        """Add comments to break WAF patterns."""
        if style == "sql":
            # Add inline comments to SQL
            return re.sub(r"(\s+)", r"/**/\1/**/", payload)
        if style == "html":
            return payload.replace("<", "<!--><")
        return payload

    @staticmethod
    def case_variation(payload: str) -> str:
        """Apply random case variation."""
        import random
        return "".join(
            c.upper() if random.random() > 0.5 else c.lower()
            for c in payload
        )

    @staticmethod
    def null_byte_inject(payload: str) -> str:
        """Inject null bytes."""
        return payload.replace(" ", "%00")

    @staticmethod
    def json_smuggle(payload: str) -> str:
        """Create JSON smuggling payload."""
        return json.dumps({"data": payload}).replace('"', '\\"')

    @staticmethod
    def header_padding(payload: str, size: int = 8000) -> str:
        """Add header padding."""
        padding = "A" * (size - len(payload))
        return padding + payload

    @staticmethod
    def chunked_encode(payload: str) -> list[str]:
        """Split payload into chunks for chunked transfer."""
        chunk_size = 10
        return [payload[i:i+chunk_size] for i in range(0, len(payload), chunk_size)]


class WAFBypass:
    """
    Main WAF bypass class that combines detection and evasion.
    """

    def __init__(self):
        self.detector = WAFDetector()
        self.encoder = PayloadEncoder()

    def detect_waf(
        self,
        status_code: int,
        headers: dict[str, str],
        body: str,
        cookies: dict[str, str] = None
    ) -> WAFDetectionResult:
        """Detect WAF from response."""
        return self.detector.detect(status_code, headers, body, cookies)

    def generate_bypass_payloads(
        self,
        original_payload: str,
        waf_type: WAFType = WAFType.UNKNOWN
    ) -> list[tuple[str, str]]:
        """
        Generate bypass payloads for a given original payload.
        Returns list of (payload, technique) tuples.
        """
        payloads = []

        # URL encoding variations
        payloads.append((
            self.encoder.url_encode(original_payload),
            "URL encoded"
        ))
        payloads.append((
            self.encoder.url_encode(original_payload, double=True),
            "Double URL encoded"
        ))

        # Unicode normalization
        payloads.append((
            self.encoder.unicode_normalize(original_payload),
            "Unicode normalized"
        ))

        # Hex encoding
        payloads.append((
            self.encoder.hex_encode(original_payload),
            "Hex encoded"
        ))

        # Case variation
        payloads.append((
            self.encoder.case_variation(original_payload),
            "Case variation"
        ))

        # Comment injection (for SQL-like payloads)
        if any(kw in original_payload.lower() for kw in ["select", "union", "insert", "update"]):
            payloads.append((
                self.encoder.add_comments(original_payload, "sql"),
                "SQL comment injection"
            ))

        # Null byte injection
        payloads.append((
            self.encoder.null_byte_inject(original_payload),
            "Null byte injection"
        ))

        # WAF-specific techniques
        if waf_type == WAFType.AKAMAI:
            payloads.append((
                self.encoder.json_smuggle(original_payload),
                "JSON smuggling (Akamai)"
            ))

        return payloads

    def get_bypass_headers(self, waf_type: WAFType = WAFType.UNKNOWN) -> dict[str, str]:
        """Get headers that may help bypass WAF."""
        headers = {
            "X-Originating-IP": "127.0.0.1",
            "X-Forwarded-For": "127.0.0.1",
            "X-Remote-IP": "127.0.0.1",
            "X-Remote-Addr": "127.0.0.1",
            "X-Client-IP": "127.0.0.1",
            "X-Real-IP": "127.0.0.1",
            "X-Custom-IP-Authorization": "127.0.0.1",
            "True-Client-IP": "127.0.0.1",
        }

        if waf_type == WAFType.CLOUDFLARE:
            headers["CF-Connecting-IP"] = "127.0.0.1"

        return headers


# Global instance
_waf_bypass: WAFBypass | None = None


def get_waf_bypass() -> WAFBypass:
    """Get the global WAFBypass instance."""
    global _waf_bypass
    if _waf_bypass is None:
        _waf_bypass = WAFBypass()
    return _waf_bypass


def detect_waf(status_code: int, headers: dict, body: str) -> WAFDetectionResult:
    """Convenience function to detect WAF."""
    return get_waf_bypass().detect_waf(status_code, headers, body)


def generate_bypasses(payload: str) -> list[tuple[str, str]]:
    """Convenience function to generate bypass payloads."""
    return get_waf_bypass().generate_bypass_payloads(payload)



from pathlib import Path

from jinja2 import Environment


def get_available_prompt_modules() -> dict[str, list[str]]:
    modules_dir = Path(__file__).parent
    available_modules = {}

    for category_dir in modules_dir.iterdir():
        if category_dir.is_dir() and not category_dir.name.startswith("__"):
            category_name = category_dir.name
            modules = []

            for file_path in category_dir.glob("*.jinja"):
                module_name = file_path.stem
                modules.append(module_name)

            if modules:
                available_modules[category_name] = sorted(modules)

    return available_modules


def get_all_module_names() -> set[str]:
    all_modules = set()
    for category_modules in get_available_prompt_modules().values():
        all_modules.update(category_modules)
    return all_modules


def auto_detect_modules(target: str, instruction: str = "") -> list[str]:
    """
    Automatically detect and return relevant prompt modules based on target and instruction.
    
    Args:
        target: The target URL or domain
        instruction: The user's instruction
        
    Returns:
        List of auto-detected module names
    """
    try:
        from exaaiagnt.prompts.auto_loader import detect_modules_from_target
        return detect_modules_from_target(target, instruction)
    except ImportError:
        return []


def get_smart_modules(target: str, instruction: str = "", user_modules: list[str] | None = None) -> list[str]:
    """
    Get final list of modules combining user-specified and auto-detected modules.
    
    Args:
        target: The target URL or domain
        instruction: The user's instruction
        user_modules: Modules explicitly specified by user
        
    Returns:
        Combined list of modules (user + auto-detected, max 5)
    """
    final_modules = set(user_modules or [])

    # Auto-detect if user didn't specify modules
    if not user_modules:
        auto_modules = auto_detect_modules(target, instruction)
        final_modules.update(auto_modules)

    # Validate and limit to 5
    available = get_all_module_names()
    valid_modules = [m for m in final_modules if m in available]

    return valid_modules[:5]


def validate_module_names(module_names: list[str]) -> dict[str, list[str]]:
    available_modules = get_all_module_names()
    valid_modules = []
    invalid_modules = []

    for module_name in module_names:
        if module_name in available_modules:
            valid_modules.append(module_name)
        else:
            invalid_modules.append(module_name)

    return {"valid": valid_modules, "invalid": invalid_modules}


def generate_modules_description() -> str:
    available_modules = get_available_prompt_modules()

    if not available_modules:
        return "No prompt modules available"

    all_module_names = get_all_module_names()

    if not all_module_names:
        return "No prompt modules available"

    sorted_modules = sorted(all_module_names)
    modules_str = ", ".join(sorted_modules)

    description = (
        f"List of prompt modules to load for this agent (max 5). Available modules: {modules_str}. "
    )

    example_modules = sorted_modules[:2]
    if example_modules:
        example = f"Example: {', '.join(example_modules)} for specialized agent"
        description += example

    return description


def load_prompt_modules(module_names: list[str], jinja_env: Environment) -> dict[str, str]:
    import logging

    logger = logging.getLogger(__name__)
    module_content = {}
    prompts_dir = Path(__file__).parent

    available_modules = get_available_prompt_modules()

    for module_name in module_names:
        try:
            module_path = None

            if "/" in module_name:
                module_path = f"{module_name}.jinja"
            else:
                for category, modules in available_modules.items():
                    if module_name in modules:
                        module_path = f"{category}/{module_name}.jinja"
                        break

                if not module_path:
                    root_candidate = f"{module_name}.jinja"
                    if (prompts_dir / root_candidate).exists():
                        module_path = root_candidate

            if module_path and (prompts_dir / module_path).exists():
                template = jinja_env.get_template(module_path)
                var_name = module_name.split("/")[-1]
                module_content[var_name] = template.render()
                logger.info(f"Loaded prompt module: {module_name} -> {var_name}")
            else:
                logger.warning(f"Prompt module not found: {module_name}")

        except (FileNotFoundError, OSError, ValueError) as e:
            logger.warning(f"Failed to load prompt module {module_name}: {e}")

    return module_content



import argparse
import asyncio
import atexit
import logging
import random
import signal
import sys
import threading
from collections.abc import Callable
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as pkg_version
from typing import TYPE_CHECKING, Any, ClassVar, cast


if TYPE_CHECKING:
    from textual.timer import Timer

from rich.align import Align
from rich.console import Group
from rich.markup import escape as rich_escape
from rich.panel import Panel
from rich.style import Style
from rich.text import Text
from textual import events, on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Grid, Horizontal, Vertical, VerticalScroll
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.widgets import Button, Label, Static, TextArea, Tree
from textual.widgets.tree import TreeNode

from exaaiagnt.agents.ExaaiAgent import ExaaiAgent
from exaaiagnt.interface.utils import build_live_stats_text
from exaaiagnt.llm.config import LLMConfig
from exaaiagnt.telemetry.tracer import Tracer, set_global_tracer


def escape_markup(text: str) -> str:
    return cast("str", rich_escape(text))


def get_package_version() -> str:
    try:
        return pkg_version("exaai-agent")
    except PackageNotFoundError:
        # Fallback version if package not installed
        return "2.2.2"


class ChatTextArea(TextArea):  # type: ignore[misc]
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._app_reference: ExaaiTUIApp | None = None

    def set_app_reference(self, app: "ExaaiTUIApp") -> None:
        self._app_reference = app

    def _on_key(self, event: events.Key) -> None:
        if event.key == "enter" and self._app_reference:
            text_content = str(self.text)  # type: ignore[has-type]
            message = text_content.strip()
            if message:
                self.text = ""

                self._app_reference._send_user_message(message)

                event.prevent_default()
                return

        super()._on_key(event)


class SplashScreen(Static):  # type: ignore[misc]
    # Modern color palette
    NEON_CYAN = "#00f5ff"
    NEON_PURPLE = "#bf00ff"
    NEON_PINK = "#ff00aa"
    NEON_GREEN = "#00ff88"
    NEON_ORANGE = "#ff8800"
    SOFT_WHITE = "#e0e0e0"

    # Enhanced ASCII Logo - ExaAi v2.2.2
    BANNER = r"""
    ███████╗██╗  ██╗ █████╗      █████╗ ██╗
    ██╔════╝╚██╗██╔╝██╔══██╗    ██╔══██╗██║
    █████╗   ╚███╔╝ ███████║    ███████║██║
    ██╔══╝   ██╔██╗ ██╔══██║    ██╔══██║██║
    ███████╗██╔╝ ██╗██║  ██║    ██║  ██║██║
    ╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝    ╚═╝  ╚═╝╚═╝
    """

    # Compact version for smaller screens
    BANNER_COMPACT = r"""
   ▓█████ ▒██   ██▒ ▄▄▄           ▄▄▄       ██▓
   ▓█   ▀ ▒▒ █ █ ▒░▒████▄        ▒████▄    ▓██▒
   ▒███   ░░  █   ░▒██  ▀█▄      ▒██  ▀█▄  ▒██▒
   ▒▓█  ▄  ░ █ █ ▒ ░██▄▄▄▄██     ░██▄▄▄▄██ ░██░
   ░▒████▒▒██▒ ▒██▒ ▓█   ▓██▒     ▓█   ▓██▒░██░
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._animation_step = 0
        self._animation_timer: Timer | None = None
        self._panel_static: Static | None = None
        self._version = "2.1.2"

    def compose(self) -> ComposeResult:
        self._version = get_package_version()
        self._animation_step = 0
        start_line = self._build_start_line_text(self._animation_step)
        panel = self._build_panel(start_line)

        panel_static = Static(panel, id="splash_content")
        self._panel_static = panel_static
        yield panel_static

    def on_mount(self) -> None:
        self._animation_timer = self.set_interval(0.45, self._animate_start_line)

    def on_unmount(self) -> None:
        if self._animation_timer is not None:
            self._animation_timer.stop()
            self._animation_timer = None

    def _animate_start_line(self) -> None:
        if not self._panel_static:
            return

        self._animation_step += 1
        start_line = self._build_start_line_text(self._animation_step)
        panel = self._build_panel(start_line)
        self._panel_static.update(panel)

    def _build_panel(self, start_line: Text) -> Panel:
        # Build colorful banner
        banner_text = self._build_colorful_banner()

        content = Group(
            Align.center(banner_text),
            Align.center(Text(" ")),
            Align.center(self._build_welcome_text()),
            Align.center(self._build_version_text()),
            Align.center(self._build_tagline_text()),
            Align.center(self._build_features_text()),
            Align.center(self._build_new_features_text()),
            Align.center(Text(" ")),
            Align.center(start_line.copy()),
        )

        return Panel.fit(content, border_style=self.NEON_PURPLE, padding=(1, 4))

    def _build_colorful_banner(self) -> Text:
        """Build a multi-colored gradient banner."""
        lines = self.BANNER.strip().split("\n")
        text = Text()

        # Color gradient from cyan to purple
        colors = [self.NEON_CYAN, self.NEON_GREEN, self.NEON_CYAN, self.NEON_PURPLE, self.NEON_PINK, self.NEON_PURPLE]

        for i, line in enumerate(lines):
            color = colors[i % len(colors)]
            text.append(line + "\n", style=Style(color=color, bold=True))

        return text

    def _build_welcome_text(self) -> Text:
        text = Text("⚡ Welcome to ", style=Style(color=self.SOFT_WHITE, bold=True))
        text.append("Exa", style=Style(color=self.NEON_CYAN, bold=True))
        text.append("Ai", style=Style(color=self.NEON_GREEN, bold=True))
        text.append("Agent", style=Style(color=self.NEON_PURPLE, bold=True))
        text.append(" ⚡", style=Style(color=self.SOFT_WHITE, bold=True))
        return text

    def _build_version_text(self) -> Text:
        text = Text("v", style=Style(color=self.SOFT_WHITE, dim=True))
        text.append(f"{self._version}", style=Style(color=self.NEON_GREEN, bold=True))
        text.append(" ", style=Style(color=self.SOFT_WHITE))
        text.append("• Adaptive LLM • Smart Fuzzing • WAF Bypass", style=Style(color=self.SOFT_WHITE, dim=True))
        return text

    def _build_new_features_text(self) -> Text:
        """Build new features highlight for v2.2.5."""
        text = Text("🔥 ", style=Style(color=self.NEON_ORANGE))
        text.append("NEW: ", style=Style(color=self.NEON_ORANGE, bold=True))
        text.append("Recon Engine", style=Style(color=self.NEON_PINK))
        text.append(" • ", style=Style(color=self.SOFT_WHITE, dim=True))
        text.append("K8s/Cloud Audit", style=Style(color=self.NEON_CYAN))
        text.append(" • ", style=Style(color=self.SOFT_WHITE, dim=True))
        text.append("OpenClaw Support", style=Style(color=self.NEON_GREEN))
        return text


    def _build_tagline_text(self) -> Text:
        text = Text("🔒 ", style=Style(color=self.NEON_ORANGE))
        text.append("Advanced AI-Powered Cybersecurity Agent", style=Style(color=self.SOFT_WHITE, italic=True))
        text.append(" 🔒", style=Style(color=self.NEON_ORANGE))
        return text

    def _build_features_text(self) -> Text:
        text = Text("[ ", style=Style(color=self.SOFT_WHITE, dim=True))
        text.append("Pentesting", style=Style(color=self.NEON_CYAN))
        text.append(" • ", style=Style(color=self.SOFT_WHITE, dim=True))
        text.append("Bug Bounty", style=Style(color=self.NEON_GREEN))
        text.append(" • ", style=Style(color=self.SOFT_WHITE, dim=True))
        text.append("Security", style=Style(color=self.NEON_PURPLE))
        text.append(" ]", style=Style(color=self.SOFT_WHITE, dim=True))
        return text

    def _build_start_line_text(self, phase: int) -> Text:
        emphasize = phase % 2 == 1
        base_style = Style(color=self.SOFT_WHITE, dim=not emphasize, bold=emphasize)

        # Animated dots
        dots = "." * ((phase % 3) + 1)

        text = Text("🚀 Starting ", style=base_style)
        text.append("Exa", style=Style(color=self.NEON_CYAN, bold=bool(emphasize)))
        text.append("Ai", style=Style(color=self.NEON_GREEN, bold=bool(emphasize)))
        text.append("Agent", style=Style(color=self.NEON_PURPLE, bold=bool(emphasize)))
        text.append(f" {dots}", style=base_style)

        return text


class HelpScreen(ModalScreen):  # type: ignore[misc]
    def compose(self) -> ComposeResult:
        yield Grid(
            Label("🔐 ExaAiAgent Help", id="help_title"),
            Label(
                "F1        Help\nCtrl+Q/C  Quit\nESC       Stop Agent\n"
                "Enter     Send message to agent\nTab       Switch panels\n↑/↓       Navigate tree",
                id="help_content",
            ),
            id="dialog",
        )

    def on_key(self, _event: events.Key) -> None:
        self.app.pop_screen()


class StopAgentScreen(ModalScreen):  # type: ignore[misc]
    def __init__(self, agent_name: str, agent_id: str):
        super().__init__()
        self.agent_name = agent_name
        self.agent_id = agent_id

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"🛑 Stop '{self.agent_name}'?", id="stop_agent_title"),
            Grid(
                Button("Yes", variant="error", id="stop_agent"),
                Button("No", variant="default", id="cancel_stop"),
                id="stop_agent_buttons",
            ),
            id="stop_agent_dialog",
        )

    def on_mount(self) -> None:
        cancel_button = self.query_one("#cancel_stop", Button)
        cancel_button.focus()

    def on_key(self, event: events.Key) -> None:
        if event.key in ("left", "right", "up", "down"):
            focused = self.focused

            if focused and focused.id == "stop_agent":
                cancel_button = self.query_one("#cancel_stop", Button)
                cancel_button.focus()
            else:
                stop_button = self.query_one("#stop_agent", Button)
                stop_button.focus()

            event.prevent_default()
        elif event.key == "enter":
            focused = self.focused
            if focused and isinstance(focused, Button):
                focused.press()
            event.prevent_default()
        elif event.key == "escape":
            self.app.pop_screen()
            event.prevent_default()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "stop_agent":
            self.app.action_confirm_stop_agent(self.agent_id)
        else:
            self.app.pop_screen()


class QuitScreen(ModalScreen):  # type: ignore[misc]
    def compose(self) -> ComposeResult:
        yield Grid(
            Label("🦉 Quit ExaaiAgnt? ", id="quit_title"),
            Grid(
                Button("Yes", variant="error", id="quit"),
                Button("No", variant="default", id="cancel"),
                id="quit_buttons",
            ),
            id="quit_dialog",
        )

    def on_mount(self) -> None:
        cancel_button = self.query_one("#cancel", Button)
        cancel_button.focus()

    def on_key(self, event: events.Key) -> None:
        if event.key in ("left", "right", "up", "down"):
            focused = self.focused

            if focused and focused.id == "quit":
                cancel_button = self.query_one("#cancel", Button)
                cancel_button.focus()
            else:
                quit_button = self.query_one("#quit", Button)
                quit_button.focus()

            event.prevent_default()
        elif event.key == "enter":
            focused = self.focused
            if focused and isinstance(focused, Button):
                focused.press()
            event.prevent_default()
        elif event.key == "escape":
            self.app.pop_screen()
            event.prevent_default()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "quit":
            self.app.action_custom_quit()
        else:
            self.app.pop_screen()


class ExaaiTUIApp(App):  # type: ignore[misc]
    CSS_PATH = "assets/tui_styles.tcss"

    selected_agent_id: reactive[str | None] = reactive(default=None)
    show_splash: reactive[bool] = reactive(default=True)

    BINDINGS: ClassVar[list[Binding]] = [
        Binding("f1", "toggle_help", "Help", priority=True),
        Binding("ctrl+q", "request_quit", "Quit", priority=True),
        Binding("ctrl+c", "request_quit", "Quit", priority=True),
        Binding("escape", "stop_selected_agent", "Stop Agent", priority=True),
    ]

    def __init__(self, args: argparse.Namespace):
        super().__init__()
        self.args = args
        self.scan_config = self._build_scan_config(args)
        self.agent_config = self._build_agent_config(args)

        self.tracer = Tracer(self.scan_config["run_name"])
        self.tracer.set_scan_config(self.scan_config)
        set_global_tracer(self.tracer)

        self.agent_nodes: dict[str, TreeNode] = {}

        self._displayed_agents: set[str] = set()
        self._displayed_events: list[str] = []

        self._scan_thread: threading.Thread | None = None
        self._scan_stop_event = threading.Event()
        self._scan_completed = threading.Event()

        self._action_verbs = [
            "Generating",
            "Scanning",
            "Analyzing",
            "Probing",
            "Hacking",
            "Testing",
            "Exploiting",
            "Investigating",
        ]
        self._agent_verbs: dict[str, str] = {}  # agent_id -> current_verb
        self._agent_verb_timers: dict[str, Any] = {}  # agent_id -> timer
        self._agent_dot_states: dict[str, int] = {}  # agent_id -> dot_count (0-3)
        self._dot_animation_timer: Any | None = None
        self._last_runtime_error: str | None = None

        self._setup_cleanup_handlers()

    def _build_scan_config(self, args: argparse.Namespace) -> dict[str, Any]:
        return {
            "scan_id": args.run_name,
            "targets": args.targets_info,
            "user_instructions": args.instruction or "",
            "run_name": args.run_name,
        }

    def _build_agent_config(self, args: argparse.Namespace) -> dict[str, Any]:
        prompt_modules = getattr(args, "resolved_prompt_modules", None)

        llm_config = LLMConfig(prompt_modules=prompt_modules)

        config = {
            "llm_config": llm_config,
            "max_iterations": 300,
        }

        if getattr(args, "local_sources", None):
            config["local_sources"] = args.local_sources

        return config

    def _setup_cleanup_handlers(self) -> None:
        def cleanup_on_exit() -> None:
            self.tracer.cleanup()

        def signal_handler(_signum: int, _frame: Any) -> None:
            self.tracer.cleanup()
            sys.exit(0)

        atexit.register(cleanup_on_exit)
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        if hasattr(signal, "SIGHUP"):
            signal.signal(signal.SIGHUP, signal_handler)

    def compose(self) -> ComposeResult:
        if self.show_splash:
            yield SplashScreen(id="splash_screen")

    def watch_show_splash(self, show_splash: bool) -> None:
        if not show_splash and self.is_mounted:
            try:
                splash = self.query_one("#splash_screen")
                splash.remove()
            except ValueError:
                pass

            main_container = Vertical(id="main_container")

            self.mount(main_container)

            content_container = Horizontal(id="content_container")
            main_container.mount(content_container)

            chat_area_container = Vertical(id="chat_area_container")

            chat_display = Static("", id="chat_display")
            chat_history = VerticalScroll(chat_display, id="chat_history")
            chat_history.can_focus = True

            status_text = Static("", id="status_text")
            keymap_indicator = Static("", id="keymap_indicator")

            agent_status_display = Horizontal(
                status_text, keymap_indicator, id="agent_status_display", classes="hidden"
            )

            chat_prompt = Static("> ", id="chat_prompt")
            chat_input = ChatTextArea(
                "",
                id="chat_input",
                show_line_numbers=False,
            )
            chat_input.set_app_reference(self)
            chat_input_container = Horizontal(chat_prompt, chat_input, id="chat_input_container")

            agents_tree = Tree("🤖 Active Agents", id="agents_tree")
            agents_tree.root.expand()
            agents_tree.show_root = False

            agents_tree.show_guide = True
            agents_tree.guide_depth = 3
            agents_tree.guide_style = "dashed"

            stats_display = Static("", id="stats_display")

            sidebar = Vertical(agents_tree, stats_display, id="sidebar")

            content_container.mount(chat_area_container)
            content_container.mount(sidebar)

            chat_area_container.mount(chat_history)
            chat_area_container.mount(agent_status_display)
            chat_area_container.mount(chat_input_container)

            self.call_after_refresh(self._focus_chat_input)

    def _focus_chat_input(self) -> None:
        if len(self.screen_stack) > 1 or self.show_splash:
            return

        if not self.is_mounted:
            return

        try:
            chat_input = self.query_one("#chat_input", ChatTextArea)
            chat_input.show_vertical_scrollbar = False
            chat_input.show_horizontal_scrollbar = False
            chat_input.focus()
        except (ValueError, Exception):
            self.call_after_refresh(self._focus_chat_input)

    def _focus_agents_tree(self) -> None:
        if len(self.screen_stack) > 1 or self.show_splash:
            return

        if not self.is_mounted:
            return

        try:
            agents_tree = self.query_one("#agents_tree", Tree)
            agents_tree.focus()

            if agents_tree.root.children:
                first_node = agents_tree.root.children[0]
                agents_tree.select_node(first_node)
        except (ValueError, Exception):
            self.call_after_refresh(self._focus_agents_tree)

    def on_mount(self) -> None:
        self.title = "exaaiagnt"

        self.set_timer(4.5, self._hide_splash_screen)

    def _hide_splash_screen(self) -> None:
        self.show_splash = False

        self._start_scan_thread()

        self.set_interval(0.5, self._update_ui_from_tracer)

    def _update_ui_from_tracer(self) -> None:
        if self.show_splash:
            return

        if len(self.screen_stack) > 1:
            return

        if not self.is_mounted:
            return

        try:
            chat_history = self.query_one("#chat_history", VerticalScroll)
            agents_tree = self.query_one("#agents_tree", Tree)

            if not self._is_widget_safe(chat_history) or not self._is_widget_safe(agents_tree):
                return
        except (ValueError, Exception):
            return

        agent_updates = False
        for agent_id, agent_data in self.tracer.agents.items():
            if agent_id not in self._displayed_agents:
                self._add_agent_node(agent_data)
                self._displayed_agents.add(agent_id)
                agent_updates = True
            elif self._update_agent_node(agent_id, agent_data):
                agent_updates = True

        if agent_updates:
            self._expand_all_agent_nodes()

        self._update_chat_view()

        self._update_agent_status_display()

        self._update_stats_display()

    def _update_agent_node(self, agent_id: str, agent_data: dict[str, Any]) -> bool:
        if agent_id not in self.agent_nodes:
            return False

        try:
            agent_node = self.agent_nodes[agent_id]
            agent_name_raw = agent_data.get("name", "Agent")
            status = agent_data.get("status", "running")

            status_indicators = {
                "running": "🟢",
                "waiting": "⏸️",
                "completed": "✅",
                "finished": "✅",
                "failed": "❌",
                "error": "❌",
                "stopped": "⏹️",
                "stopping": "⏸️",
                "llm_failed": "🔴",
            }

            status_icon = status_indicators.get(status, "🔵")
            agent_name = f"{status_icon} {escape_markup(agent_name_raw)}"

            if status == "running":
                self._start_agent_verb_timer(agent_id)
            elif status == "waiting":
                self._stop_agent_verb_timer(agent_id)
            else:
                self._stop_agent_verb_timer(agent_id)

            if agent_node.label != agent_name:
                agent_node.set_label(agent_name)
                return True

        except (KeyError, AttributeError, ValueError) as e:
            import logging

            logging.warning(f"Failed to update agent node label: {e}")

        return False

    def _update_chat_view(self) -> None:
        if len(self.screen_stack) > 1 or self.show_splash:
            return

        if not self.is_mounted:
            return

        try:
            chat_history = self.query_one("#chat_history", VerticalScroll)
        except (ValueError, Exception):
            return

        if not self._is_widget_safe(chat_history):
            return

        try:
            is_at_bottom = chat_history.scroll_y >= chat_history.max_scroll_y
        except (AttributeError, ValueError):
            is_at_bottom = True

        if not self.selected_agent_id:
            content, css_class = self._get_chat_placeholder_content(
                "Select an agent from the tree to see its activity.", "placeholder-no-agent"
            )
        else:
            events = self._gather_agent_events(self.selected_agent_id)
            if not events:
                content, css_class = self._get_chat_placeholder_content(
                    "Starting agent...", "placeholder-no-activity"
                )
            else:
                current_event_ids = [e["id"] for e in events]
                if current_event_ids == self._displayed_events:
                    return
                content = self._get_rendered_events_content(events)
                css_class = "chat-content"
                self._displayed_events = current_event_ids

        chat_display = self.query_one("#chat_display", Static)
        self._update_static_content_safe(chat_display, content)

        chat_display.set_classes(css_class)

        if is_at_bottom:
            self.call_later(chat_history.scroll_end, animate=False)

    def _get_chat_placeholder_content(
        self, message: str, placeholder_class: str
    ) -> tuple[str, str]:
        self._displayed_events = [placeholder_class]
        return message, f"chat-placeholder {placeholder_class}"

    def _get_rendered_events_content(self, events: list[dict[str, Any]]) -> str:
        if not events:
            return ""

        content_lines = []
        for event in events:
            if event["type"] == "chat":
                chat_content = self._render_chat_content(event["data"])
                if chat_content:
                    content_lines.append(chat_content)
            elif event["type"] == "tool":
                tool_content = self._render_tool_content_simple(event["data"])
                if tool_content:
                    content_lines.append(tool_content)

        return "\n\n".join(content_lines)

    def _update_agent_status_display(self) -> None:
        try:
            status_display = self.query_one("#agent_status_display", Horizontal)
            status_text = self.query_one("#status_text", Static)
            keymap_indicator = self.query_one("#keymap_indicator", Static)
        except (ValueError, Exception):
            return

        widgets = [status_display, status_text, keymap_indicator]
        if not all(self._is_widget_safe(w) for w in widgets):
            return

        if not self.selected_agent_id:
            self._safe_widget_operation(status_display.add_class, "hidden")
            return

        try:
            agent_data = self.tracer.agents[self.selected_agent_id]
            status = agent_data.get("status", "running")

            if status == "stopping":
                self._safe_widget_operation(status_text.update, "Agent stopping...")
                self._safe_widget_operation(keymap_indicator.update, "")
                self._safe_widget_operation(status_display.remove_class, "hidden")
            elif status == "stopped":
                self._safe_widget_operation(status_text.update, "Agent stopped")
                self._safe_widget_operation(keymap_indicator.update, "")
                self._safe_widget_operation(status_display.remove_class, "hidden")
            elif status in {"completed", "finished"}:
                self._safe_widget_operation(status_text.update, "Agent completed")
                self._safe_widget_operation(keymap_indicator.update, "")
                self._safe_widget_operation(status_display.remove_class, "hidden")
            elif status in {"llm_failed", "failed", "error"}:
                error_msg = agent_data.get("error_message", "")
                display_msg = (
                    f"[red]{escape_markup(error_msg)}[/red]"
                    if error_msg
                    else "[red]LLM request failed[/red]"
                )
                self._safe_widget_operation(status_text.update, display_msg)
                self._safe_widget_operation(
                    keymap_indicator.update, "[dim]Send message to retry[/dim]"
                )
                self._safe_widget_operation(status_display.remove_class, "hidden")
                self._stop_dot_animation()
            elif status == "waiting":
                animated_text = self._get_animated_waiting_text(self.selected_agent_id)
                self._safe_widget_operation(status_text.update, animated_text)
                self._safe_widget_operation(
                    keymap_indicator.update, "[dim]Send message to resume[/dim]"
                )
                self._safe_widget_operation(status_display.remove_class, "hidden")
                self._start_dot_animation()
            elif status == "running":
                current_verb = self._get_agent_verb(self.selected_agent_id)
                animated_text = self._get_animated_verb_text(self.selected_agent_id, current_verb)
                self._safe_widget_operation(status_text.update, animated_text)
                self._safe_widget_operation(
                    keymap_indicator.update, "[dim]ESC to stop | CTRL-C to quit and save[/dim]"
                )
                self._safe_widget_operation(status_display.remove_class, "hidden")
                self._start_dot_animation()
            else:
                self._safe_widget_operation(status_display.add_class, "hidden")

        except (KeyError, Exception):
            self._safe_widget_operation(status_display.add_class, "hidden")

    def _update_stats_display(self) -> None:
        try:
            stats_display = self.query_one("#stats_display", Static)
        except (ValueError, Exception):
            return

        if not self._is_widget_safe(stats_display):
            return

        stats_content = Text()

        stats_text = build_live_stats_text(self.tracer, self.agent_config)
        if stats_text:
            stats_content.append(stats_text)

        if self._last_runtime_error:
            stats_content.append("\n")
            stats_content.append("Runtime: ", style="bold red")
            stats_content.append(self._last_runtime_error, style="red")

        from rich.panel import Panel

        stats_panel = Panel(
            stats_content,
            title="📊 Live Stats",
            title_align="left",
            border_style="#22c55e",
            padding=(0, 1),
        )

        self._safe_widget_operation(stats_display.update, stats_panel)

    def _get_agent_verb(self, agent_id: str) -> str:
        if agent_id not in self._agent_verbs:
            self._agent_verbs[agent_id] = random.choice(self._action_verbs)  # nosec B311 # noqa: S311
        return self._agent_verbs[agent_id]

    def _start_agent_verb_timer(self, agent_id: str) -> None:
        if agent_id not in self._agent_verb_timers:
            self._agent_verb_timers[agent_id] = self.set_interval(
                30.0, lambda: self._change_agent_action_verb(agent_id)
            )

    def _stop_agent_verb_timer(self, agent_id: str) -> None:
        if agent_id in self._agent_verb_timers:
            self._agent_verb_timers[agent_id].stop()
            del self._agent_verb_timers[agent_id]

    def _change_agent_action_verb(self, agent_id: str) -> None:
        if agent_id not in self._agent_verbs:
            self._agent_verbs[agent_id] = random.choice(self._action_verbs)  # nosec B311 # noqa: S311
            return

        current_verb = self._agent_verbs[agent_id]
        available_verbs = [verb for verb in self._action_verbs if verb != current_verb]
        self._agent_verbs[agent_id] = random.choice(available_verbs)  # nosec B311 # noqa: S311

        if self.selected_agent_id == agent_id:
            self._update_agent_status_display()

    def _get_animated_verb_text(self, agent_id: str, verb: str) -> str:
        if agent_id not in self._agent_dot_states:
            self._agent_dot_states[agent_id] = 0

        dot_count = self._agent_dot_states[agent_id]
        dots = "." * dot_count
        return f"{verb}{dots}"

    def _get_animated_waiting_text(self, agent_id: str) -> str:
        if agent_id not in self._agent_dot_states:
            self._agent_dot_states[agent_id] = 0

        dot_count = self._agent_dot_states[agent_id]
        dots = "." * dot_count

        return f"Waiting{dots}"

    def _start_dot_animation(self) -> None:
        if self._dot_animation_timer is None:
            self._dot_animation_timer = self.set_interval(0.6, self._animate_dots)

    def _stop_dot_animation(self) -> None:
        if self._dot_animation_timer is not None:
            self._dot_animation_timer.stop()
            self._dot_animation_timer = None

    def _animate_dots(self) -> None:
        has_active_agents = False

        for agent_id, agent_data in self.tracer.agents.items():
            status = agent_data.get("status", "running")
            if status in ["running", "waiting"]:
                has_active_agents = True
                current_dots = self._agent_dot_states.get(agent_id, 0)
                self._agent_dot_states[agent_id] = (current_dots + 1) % 4

        if (
            has_active_agents
            and self.selected_agent_id
            and self.selected_agent_id in self.tracer.agents
        ):
            selected_status = self.tracer.agents[self.selected_agent_id].get("status", "running")
            if selected_status in ["running", "waiting"]:
                self._update_agent_status_display()

        if not has_active_agents:
            self._stop_dot_animation()
            for agent_id in list(self._agent_dot_states.keys()):
                if agent_id not in self.tracer.agents or self.tracer.agents[agent_id].get(
                    "status"
                ) not in ["running", "waiting"]:
                    del self._agent_dot_states[agent_id]

    def _normalize_event_timestamp(self, raw_timestamp: Any) -> float:
        if isinstance(raw_timestamp, (int, float)):
            return float(raw_timestamp)

        if isinstance(raw_timestamp, str):
            try:
                return float(raw_timestamp)
            except ValueError:
                pass

            try:
                normalized = raw_timestamp.replace("Z", "+00:00")
                from datetime import datetime

                return datetime.fromisoformat(normalized).timestamp()
            except ValueError:
                return 0.0

        return 0.0

    def _gather_agent_events(self, agent_id: str) -> list[dict[str, Any]]:
        chat_events = [
            {
                "type": "chat",
                "timestamp": self._normalize_event_timestamp(msg.get("timestamp")),
                "id": f"chat_{msg['message_id']}",
                "data": msg,
            }
            for msg in self.tracer.chat_messages
            if msg.get("agent_id") == agent_id
        ]

        tool_events = [
            {
                "type": "tool",
                "timestamp": self._normalize_event_timestamp(tool_data.get("timestamp")),
                "id": f"tool_{exec_id}",
                "data": tool_data,
            }
            for exec_id, tool_data in self.tracer.tool_executions.items()
            if tool_data.get("agent_id") == agent_id
        ]

        events = chat_events + tool_events
        events.sort(key=lambda e: (e["timestamp"], e["id"]))
        return events

    def watch_selected_agent_id(self, _agent_id: str | None) -> None:
        if len(self.screen_stack) > 1 or self.show_splash:
            return

        if not self.is_mounted:
            return

        self._displayed_events.clear()

        self.call_later(self._update_chat_view)
        self._update_agent_status_display()

    def _record_runtime_error(self, message: str) -> None:
        self._last_runtime_error = message
        logging.error(message)

    def _start_scan_if_ready(self) -> None:
        if not self.scan_config.get("targets"):
            self._record_runtime_error("No active target. Enter a target URL/domain/path to start scanning.")
            return

        if self._scan_thread and self._scan_thread.is_alive():
            return

        self._scan_stop_event.clear()
        self._scan_completed.clear()
        self._last_runtime_error = None
        self._start_scan_thread()

    def _submit_new_target(self, raw_text: str) -> bool:
        target_text = raw_text.strip()
        if not target_text:
            self._record_runtime_error("Target cannot be empty.")
            return False

        try:
            target_type, target_dict = infer_target_type(target_text)
        except ValueError as e:
            self._record_runtime_error(f"Invalid target: {e}")
            return False

        target_info = {"type": target_type, "details": target_dict, "original": target_text}
        targets_info = [target_info]
        assign_workspace_subdirs(targets_info)

        if target_type == "repository":
            repo_url = target_dict["target_repo"]
            dest_name = target_dict.get("workspace_subdir")
            cloned_path = clone_repository(repo_url, self.scan_config["run_name"], dest_name)
            target_info["details"]["cloned_repo_path"] = cloned_path

        self.scan_config["targets"] = targets_info
        self.scan_config["user_instructions"] = ""
        self.agent_config["local_sources"] = collect_local_sources(targets_info)
        self.tracer.set_scan_config(self.scan_config)
        self._displayed_events.clear()
        self._record_runtime_error(f"Target queued: {target_text}")
        self._start_scan_if_ready()
        return True

    def _start_scan_thread(self) -> None:
        def scan_target() -> None:
            try:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)

                try:
                    agent = ExaaiAgent(self.agent_config)

                    if not self._scan_stop_event.is_set():
                        result = loop.run_until_complete(agent.execute_scan(self.scan_config))
                        if isinstance(result, dict) and not result.get("success", True):
                            self._record_runtime_error(
                                f"Scan failed: {result.get('error', 'Unknown scan failure')}"
                            )

                except (KeyboardInterrupt, asyncio.CancelledError):
                    logging.info("Scan interrupted by user")
                except (ConnectionError, TimeoutError) as e:
                    self._record_runtime_error(f"Network error during scan: {e}")
                    logging.exception("Network error during scan")
                except RuntimeError as e:
                    self._record_runtime_error(f"Runtime error during scan: {e}")
                    logging.exception("Runtime error during scan")
                except Exception as e:
                    self._record_runtime_error(f"Unexpected error during scan: {e}")
                    logging.exception("Unexpected error during scan")
                finally:
                    loop.close()
                    self._scan_completed.set()

            except Exception as e:
                self._record_runtime_error(f"Error setting up scan thread: {e}")
                logging.exception("Error setting up scan thread")
                self._scan_completed.set()

        self._scan_thread = threading.Thread(target=scan_target, daemon=True)
        self._scan_thread.start()

    def _add_agent_node(self, agent_data: dict[str, Any]) -> None:
        if len(self.screen_stack) > 1 or self.show_splash:
            return

        if not self.is_mounted:
            return

        agent_id = agent_data["id"]
        parent_id = agent_data.get("parent_id")
        status = agent_data.get("status", "running")

        try:
            agents_tree = self.query_one("#agents_tree", Tree)
        except (ValueError, Exception):
            return

        agent_name_raw = agent_data.get("name", "Agent")

        status_indicators = {
            "running": "🟢",
            "waiting": "🟡",
            "completed": "✅",
            "finished": "✅",
            "failed": "❌",
            "error": "❌",
            "stopped": "⏹️",
            "stopping": "⏸️",
            "llm_failed": "🔴",
        }

        status_icon = status_indicators.get(status, "🔵")
        agent_name = f"{status_icon} {escape_markup(agent_name_raw)}"

        if status in ["running", "waiting"]:
            self._start_agent_verb_timer(agent_id)

        try:
            if parent_id and parent_id in self.agent_nodes:
                parent_node = self.agent_nodes[parent_id]
                agent_node = parent_node.add(
                    agent_name,
                    data={"agent_id": agent_id},
                )
                parent_node.allow_expand = True
            else:
                agent_node = agents_tree.root.add(
                    agent_name,
                    data={"agent_id": agent_id},
                )

            agent_node.allow_expand = False
            agent_node.expand()
            self.agent_nodes[agent_id] = agent_node

            if len(self.agent_nodes) == 1:
                agents_tree.select_node(agent_node)
                self.selected_agent_id = agent_id

            self._reorganize_orphaned_agents(agent_id)
        except (AttributeError, ValueError, RuntimeError) as e:
            import logging

            logging.warning(f"Failed to add agent node {agent_id}: {e}")

    def _expand_all_agent_nodes(self) -> None:
        if len(self.screen_stack) > 1 or self.show_splash:
            return

        if not self.is_mounted:
            return

        try:
            agents_tree = self.query_one("#agents_tree", Tree)
            self._expand_node_recursively(agents_tree.root)
        except (ValueError, Exception):
            logging.debug("Tree not ready for expanding nodes")

    def _expand_node_recursively(self, node: TreeNode) -> None:
        if not node.is_expanded:
            node.expand()
        for child in node.children:
            self._expand_node_recursively(child)

    def _copy_node_under(self, node_to_copy: TreeNode, new_parent: TreeNode) -> None:
        agent_id = node_to_copy.data["agent_id"]
        agent_data = self.tracer.agents.get(agent_id, {})
        agent_name_raw = agent_data.get("name", "Agent")
        status = agent_data.get("status", "running")

        status_indicators = {
            "running": "🟢",
            "waiting": "🟡",
            "completed": "✅",
            "failed": "❌",
            "stopped": "⏹️",
            "stopping": "⏸️",
        }

        status_icon = status_indicators.get(status, "🔵")
        agent_name = f"{status_icon} {escape_markup(agent_name_raw)}"

        new_node = new_parent.add(
            agent_name,
            data=node_to_copy.data,
        )
        new_node.allow_expand = node_to_copy.allow_expand

        self.agent_nodes[agent_id] = new_node

        for child in node_to_copy.children:
            self._copy_node_under(child, new_node)

        if node_to_copy.is_expanded:
            new_node.expand()

    def _reorganize_orphaned_agents(self, new_parent_id: str) -> None:
        agents_to_move = []

        for agent_id, agent_data in self.tracer.agents.items():
            if (
                agent_data.get("parent_id") == new_parent_id
                and agent_id in self.agent_nodes
                and agent_id != new_parent_id
            ):
                agents_to_move.append(agent_id)

        if not agents_to_move:
            return

        parent_node = self.agent_nodes[new_parent_id]

        for child_agent_id in agents_to_move:
            if child_agent_id in self.agent_nodes:
                old_node = self.agent_nodes[child_agent_id]

                if old_node.parent is parent_node:
                    continue

                self._copy_node_under(old_node, parent_node)

                old_node.remove()

        parent_node.allow_expand = True
        self._expand_all_agent_nodes()

    def _render_chat_content(self, msg_data: dict[str, Any]) -> str:
        role = msg_data.get("role")
        content = escape_markup(msg_data.get("content", ""))

        if not content:
            return ""

        if role == "user":
            from exaaiagnt.interface.tool_components.user_message_renderer import (
                UserMessageRenderer,
            )

            return UserMessageRenderer.render_simple(content)
        return content

    def _render_tool_content_simple(self, tool_data: dict[str, Any]) -> str:
        tool_name = tool_data.get("tool_name", "Unknown Tool")
        args = tool_data.get("args", {})
        status = tool_data.get("status", "unknown")
        result = tool_data.get("result")

        tool_colors = {
            "terminal_execute": "#22c55e",
            "browser_action": "#06b6d4",
            "python_action": "#3b82f6",
            "agents_graph_action": "#fbbf24",
            "file_edit_action": "#10b981",
            "proxy_action": "#06b6d4",
            "notes_action": "#fbbf24",
            "thinking_action": "#a855f7",
            "web_search_action": "#22c55e",
            "finish_action": "#dc2626",
            "reporting_action": "#ea580c",
            "scan_start_info": "#22c55e",
            "subagent_start_info": "#22c55e",
            "llm_error_details": "#dc2626",
        }

        color = tool_colors.get(tool_name, "#737373")

        from exaaiagnt.interface.tool_components.registry import get_tool_renderer

        renderer = get_tool_renderer(tool_name)

        if renderer:
            widget = renderer.render(tool_data)
            content = str(widget.renderable)
        elif tool_name == "llm_error_details":
            lines = ["[red]✗ LLM Request Failed[/red]"]
            if args.get("details"):
                details = args["details"]
                if len(details) > 300:
                    details = details[:297] + "..."
                lines.append(f"[dim]Details:[/dim] {escape_markup(details)}")
            content = "\n".join(lines)
        else:
            status_icons = {
                "running": "[yellow]●[/yellow]",
                "completed": "[green]✓[/green]",
                "failed": "[red]✗[/red]",
                "error": "[red]✗[/red]",
            }
            status_icon = status_icons.get(status, "[dim]○[/dim]")

            lines = [f"→ Using tool [bold blue]{escape_markup(tool_name)}[/] {status_icon}"]

            if args:
                for k, v in list(args.items())[:2]:
                    str_v = str(v)
                    if len(str_v) > 80:
                        str_v = str_v[:77] + "..."
                    lines.append(f"  [dim]{k}:[/] {escape_markup(str_v)}")

            if status in ["completed", "failed", "error"] and result:
                result_str = str(result)
                if len(result_str) > 150:
                    result_str = result_str[:147] + "..."
                lines.append(f"[bold]Result:[/] {escape_markup(result_str)}")

            content = "\n".join(lines)

        lines = content.split("\n")
        bordered_lines = [f"[{color}]▍[/{color}] {line}" for line in lines]
        return "\n".join(bordered_lines)

    @on(Tree.NodeHighlighted)  # type: ignore[misc]
    def handle_tree_highlight(self, event: Tree.NodeHighlighted) -> None:
        if len(self.screen_stack) > 1 or self.show_splash:
            return

        if not self.is_mounted:
            return

        node = event.node

        try:
            agents_tree = self.query_one("#agents_tree", Tree)
        except (ValueError, Exception):
            return

        if self.focused == agents_tree and node.data:
            agent_id = node.data.get("agent_id")
            if agent_id:
                self.selected_agent_id = agent_id

    def _send_user_message(self, message: str) -> None:
        if not self.selected_agent_id:
            if self._submit_new_target(message):
                self.call_after_refresh(self._focus_chat_input)
            return

        if self.tracer:
            self.tracer.log_chat_message(
                content=message,
                role="user",
                agent_id=self.selected_agent_id,
            )

        try:
            from exaaiagnt.tools.agents_graph.agents_graph_actions import send_user_message_to_agent

            result = send_user_message_to_agent(self.selected_agent_id, message)
            if not result.get("success", False):
                error_message = result.get("error", "unknown error")
                self._record_runtime_error(
                    f"Failed to send message to agent {self.selected_agent_id}: {error_message}"
                )
                logging.warning(
                    "Failed to send message to agent %s: %s",
                    self.selected_agent_id,
                    error_message,
                )

        except (ImportError, AttributeError) as e:
            self._record_runtime_error(
                f"Failed to send message to agent {self.selected_agent_id}: {e}"
            )
            logging.warning(f"Failed to send message to agent {self.selected_agent_id}: {e}")

        self._displayed_events.clear()
        self._update_chat_view()

        self.call_after_refresh(self._focus_chat_input)

    def _get_agent_name(self, agent_id: str) -> str:
        try:
            if self.tracer and agent_id in self.tracer.agents:
                agent_name = self.tracer.agents[agent_id].get("name")
                if isinstance(agent_name, str):
                    return agent_name
        except (KeyError, AttributeError) as e:
            logging.warning(f"Could not retrieve agent name for {agent_id}: {e}")
        return "Unknown Agent"

    def action_toggle_help(self) -> None:
        if self.show_splash or not self.is_mounted:
            return

        try:
            self.query_one("#main_container")
        except (ValueError, Exception):
            return

        if isinstance(self.screen, HelpScreen):
            self.pop_screen()
            return

        if len(self.screen_stack) > 1:
            return

        self.push_screen(HelpScreen())

    def action_request_quit(self) -> None:
        if self.show_splash or not self.is_mounted:
            self.action_custom_quit()
            return

        if len(self.screen_stack) > 1:
            return

        try:
            self.query_one("#main_container")
        except (ValueError, Exception):
            self.action_custom_quit()
            return

        self.push_screen(QuitScreen())

    def action_stop_selected_agent(self) -> None:
        if (
            self.show_splash
            or not self.is_mounted
            or len(self.screen_stack) > 1
            or not self.selected_agent_id
        ):
            return

        agent_name, should_stop = self._validate_agent_for_stopping()
        if not should_stop:
            return

        try:
            self.query_one("#main_container")
        except (ValueError, Exception):
            return

        self.push_screen(StopAgentScreen(agent_name, self.selected_agent_id))

    def _validate_agent_for_stopping(self) -> tuple[str, bool]:
        agent_name = "Unknown Agent"

        try:
            if self.tracer and self.selected_agent_id in self.tracer.agents:
                agent_data = self.tracer.agents[self.selected_agent_id]
                agent_name = agent_data.get("name", "Unknown Agent")

                agent_status = agent_data.get("status", "running")
                if agent_status not in ["running"]:
                    return agent_name, False

                agent_events = self._gather_agent_events(self.selected_agent_id)
                if not agent_events:
                    return agent_name, False

                return agent_name, True

        except (KeyError, AttributeError, ValueError) as e:
            import logging

            logging.warning(f"Failed to gather agent events: {e}")

        return agent_name, False

    def action_confirm_stop_agent(self, agent_id: str) -> None:
        self.pop_screen()

        try:
            from exaaiagnt.tools.agents_graph.agents_graph_actions import stop_agent

            result = stop_agent(agent_id)

            import logging

            if result.get("success"):
                logging.info(f"Stop request sent to agent: {result.get('message', 'Unknown')}")
            else:
                logging.warning(f"Failed to stop agent: {result.get('error', 'Unknown error')}")

        except Exception:
            import logging

            logging.exception(f"Failed to stop agent {agent_id}")

    def action_custom_quit(self) -> None:
        for agent_id in list(self._agent_verb_timers.keys()):
            self._stop_agent_verb_timer(agent_id)

        if self._scan_thread and self._scan_thread.is_alive():
            self._scan_stop_event.set()

            self._scan_thread.join(timeout=1.0)

        self.tracer.cleanup()

        self.exit()

    def _is_widget_safe(self, widget: Any) -> bool:
        try:
            _ = widget.screen
        except (AttributeError, ValueError, Exception):
            return False
        else:
            return bool(widget.is_mounted)

    def _safe_widget_operation(
        self, operation: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> bool:
        try:
            operation(*args, **kwargs)
        except (AttributeError, ValueError, Exception):
            return False
        else:
            return True

    def _update_static_content_safe(self, widget: Static, content: str) -> None:
        try:
            widget.update(content)
        except Exception:  # noqa: BLE001
            try:
                safe_text = Text.from_markup(content)
                widget.update(safe_text)
            except Exception:  # noqa: BLE001
                import re

                plain_text = re.sub(r"\[.*?\]", "", content)
                widget.update(plain_text)


async def run_tui(args: argparse.Namespace) -> None:
    """Run ExaaiAgnt in interactive TUI mode with textual."""
    app = ExaaiTUIApp(args)
    await app.run_async()



"""
Tool Manager - Manages tool lifecycle with process isolation.
Provides monitoring, restart, and graceful degradation capabilities.
"""

import asyncio
import logging
import os
import time
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional


logger = logging.getLogger(__name__)


class ToolStatus(Enum):
    """Tool lifecycle states."""
    IDLE = "idle"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    TIMEOUT = "timeout"
    KILLED = "killed"


@dataclass
class ToolExecution:
    """Represents a single tool execution."""
    tool_id: str
    tool_name: str
    args: dict[str, Any] = field(default_factory=dict)
    status: ToolStatus = ToolStatus.IDLE
    start_time: float = 0
    end_time: float = 0
    process_id: int | None = None
    output: Any = None
    error: str | None = None
    timeout: float = 300  # 5 minutes default
    retries: int = 0
    max_retries: int = 3


class ToolManager:
    """
    Manages tool lifecycle with process isolation.
    
    Features:
    - Process isolation per tool
    - Tool health monitoring
    - Automatic restart on failure
    - Graceful degradation
    - Kill/restart capabilities
    """

    DEFAULT_TIMEOUT = 300  # 5 minutes
    MAX_CONCURRENT_TOOLS = 5
    HEALTH_CHECK_INTERVAL = 10

    _instance: Optional["ToolManager"] = None
    _lock = __import__("threading").Lock()

    def __new__(cls) -> "ToolManager":
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
                cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        if self._initialized:
            return

        self._executions: dict[str, ToolExecution] = {}
        self._running_processes: dict[str, asyncio.subprocess.Process] = {}
        self._tool_semaphore = asyncio.Semaphore(self.MAX_CONCURRENT_TOOLS)
        self._monitor_task: asyncio.Task | None = None
        self._running = False
        self._executor = ThreadPoolExecutor(max_workers=10)

        # Callbacks
        self._on_failure_callbacks: list[Callable] = []
        self._on_complete_callbacks: list[Callable] = []

        # Statistics
        self._stats = {
            "total_executions": 0,
            "successful": 0,
            "failed": 0,
            "timeouts": 0,
            "retries": 0,
        }

        self._initialized = True
        logger.info("ToolManager initialized")

    async def shutdown(self):
        """Cleanup resources."""
        await self.kill_all()
        self._executor.shutdown(wait=False)
        self._running = False
        logger.info("ToolManager shutdown complete")

    async def execute(
        self,
        tool_name: str,
        command: str,
        args: dict[str, Any] = None,
        timeout: float = None,
        cwd: str = None,
        env: dict[str, str] = None,
        retry_on_failure: bool = True
    ) -> ToolExecution:
        """
        Execute a tool with process isolation.
        
        Args:
            tool_name: Name of the tool
            command: Command to execute
            args: Arguments for the tool
            timeout: Execution timeout in seconds
            cwd: Working directory
            env: Environment variables
            retry_on_failure: Whether to retry on failure
            
        Returns:
            ToolExecution with results
        """
        tool_id = f"{tool_name}_{int(time.time() * 1000)}"
        execution = ToolExecution(
            tool_id=tool_id,
            tool_name=tool_name,
            args=args or {},
            timeout=timeout or self.DEFAULT_TIMEOUT,
        )

        self._executions[tool_id] = execution
        self._stats["total_executions"] += 1

        async with self._tool_semaphore:
            try:
                result = await self._run_with_timeout(
                    execution, command, cwd, env
                )

                if result.status == ToolStatus.FAILED and retry_on_failure:
                    result = await self._retry_execution(
                        execution, command, cwd, env
                    )

                return result

            except Exception as e:
                execution.status = ToolStatus.FAILED
                execution.error = str(e)
                self._stats["failed"] += 1
                logger.error(f"Tool {tool_name} failed: {e}")

                self._notify_failure(execution)
                return execution

    async def _run_with_timeout(
        self,
        execution: ToolExecution,
        command: str,
        cwd: str = None,
        env: dict[str, str] = None
    ) -> ToolExecution:
        """Run command with timeout handling."""
        execution.status = ToolStatus.RUNNING
        execution.start_time = time.time()

        # Prepare environment
        process_env = os.environ.copy()
        if env:
            process_env.update(env)

        try:
            # Create subprocess
            process = await asyncio.create_subprocess_shell(
                command,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                cwd=cwd,
                env=process_env,
            )

            execution.process_id = process.pid
            self._running_processes[execution.tool_id] = process

            try:
                # Wait with timeout
                stdout, stderr = await asyncio.wait_for(
                    process.communicate(),
                    timeout=execution.timeout
                )

                execution.end_time = time.time()
                execution.output = stdout.decode("utf-8", errors="replace")

                if process.returncode == 0:
                    execution.status = ToolStatus.COMPLETED
                    self._stats["successful"] += 1
                else:
                    execution.status = ToolStatus.FAILED
                    execution.error = stderr.decode("utf-8", errors="replace")
                    self._stats["failed"] += 1

            except TimeoutError:
                execution.status = ToolStatus.TIMEOUT
                execution.error = f"Timeout after {execution.timeout}s"
                execution.end_time = time.time()
                self._stats["timeouts"] += 1

                # Kill the process
                await self._kill_process(execution.tool_id)

            finally:
                # Cleanup
                if execution.tool_id in self._running_processes:
                    del self._running_processes[execution.tool_id]

        except Exception as e:
            execution.status = ToolStatus.FAILED
            execution.error = str(e)
            execution.end_time = time.time()
            self._stats["failed"] += 1

        self._notify_complete(execution)
        return execution

    async def _retry_execution(
        self,
        execution: ToolExecution,
        command: str,
        cwd: str = None,
        env: dict[str, str] = None
    ) -> ToolExecution:
        """Retry failed execution."""
        while execution.retries < execution.max_retries:
            execution.retries += 1
            self._stats["retries"] += 1

            logger.info(
                f"Retrying {execution.tool_name} "
                f"(attempt {execution.retries}/{execution.max_retries})"
            )

            # Wait before retry (exponential backoff)
            await asyncio.sleep(min(2 ** execution.retries, 30))

            # Reset for retry
            execution.status = ToolStatus.IDLE
            execution.error = None

            result = await self._run_with_timeout(execution, command, cwd, env)

            if result.status == ToolStatus.COMPLETED:
                return result

        logger.error(
            f"Tool {execution.tool_name} failed after "
            f"{execution.max_retries} retries"
        )
        return execution

    async def _kill_process(self, tool_id: str) -> bool:
        """Kill a running process."""
        if tool_id not in self._running_processes:
            return False

        process = self._running_processes[tool_id]

        try:
            # Try graceful termination first
            process.terminate()

            try:
                await asyncio.wait_for(process.wait(), timeout=5)
            except TimeoutError:
                # Force kill
                process.kill()
                await process.wait()

            logger.info(f"Killed process for tool {tool_id}")

            if tool_id in self._executions:
                self._executions[tool_id].status = ToolStatus.KILLED

            return True

        except Exception as e:
            logger.error(f"Failed to kill process {tool_id}: {e}")
            return False
        finally:
            if tool_id in self._running_processes:
                del self._running_processes[tool_id]

    async def kill_tool(self, tool_id: str) -> bool:
        """Public method to kill a tool by ID."""
        return await self._kill_process(tool_id)

    async def kill_all(self) -> int:
        """Kill all running tools."""
        killed = 0
        for tool_id in list(self._running_processes.keys()):
            if await self._kill_process(tool_id):
                killed += 1
        return killed

    def get_execution(self, tool_id: str) -> ToolExecution | None:
        """Get execution by ID."""
        return self._executions.get(tool_id)

    def get_running_tools(self) -> list[ToolExecution]:
        """Get list of currently running tools."""
        return [
            self._executions[tid]
            for tid in self._running_processes.keys()
            if tid in self._executions
        ]

    def get_stats(self) -> dict[str, int]:
        """Get execution statistics."""
        return dict(self._stats)

    def on_failure(self, callback: Callable[[ToolExecution], None]) -> None:
        """Register callback for tool failures."""
        self._on_failure_callbacks.append(callback)

    def on_complete(self, callback: Callable[[ToolExecution], None]) -> None:
        """Register callback for tool completion."""
        self._on_complete_callbacks.append(callback)

    def _notify_failure(self, execution: ToolExecution) -> None:
        """Notify failure callbacks."""
        for callback in self._on_failure_callbacks:
            try:
                callback(execution)
            except Exception as e:
                logger.error(f"Failure callback error: {e}")

    def _notify_complete(self, execution: ToolExecution) -> None:
        """Notify completion callbacks."""
        for callback in self._on_complete_callbacks:
            try:
                callback(execution)
            except Exception as e:
                logger.error(f"Complete callback error: {e}")

    async def start_monitoring(self) -> None:
        """Start the tool monitoring loop."""
        if self._running:
            return

        self._running = True
        self._monitor_task = asyncio.create_task(self._monitoring_loop())
        logger.info("Tool monitoring started")

    async def stop_monitoring(self) -> None:
        """Stop the tool monitoring loop."""
        self._running = False
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass
        logger.info("Tool monitoring stopped")

    async def _monitoring_loop(self) -> None:
        """Main monitoring loop for tools."""
        while self._running:
            try:
                await self._check_tool_health()
                await asyncio.sleep(self.HEALTH_CHECK_INTERVAL)
            except asyncio.CancelledError:
                break
            except Exception as e:
                logger.error(f"Tool monitoring error: {e}")
                await asyncio.sleep(1)

    async def _check_tool_health(self) -> None:
        """Check health of all running tools."""
        now = time.time()

        for tool_id, execution in list(self._executions.items()):
            if execution.status != ToolStatus.RUNNING:
                continue

            # Check if process still exists
            if tool_id in self._running_processes:
                process = self._running_processes[tool_id]
                if process.returncode is not None:
                    # Process has ended
                    execution.status = ToolStatus.COMPLETED if process.returncode == 0 else ToolStatus.FAILED
                    execution.end_time = now
                    del self._running_processes[tool_id]

            # Check timeout
            if execution.start_time > 0:
                elapsed = now - execution.start_time
                if elapsed > execution.timeout:
                    logger.warning(
                        f"Tool {execution.tool_name} ({tool_id}) timed out "
                        f"after {elapsed:.0f}s"
                    )
                    await self._kill_process(tool_id)


# Global instance
def get_tool_manager() -> ToolManager:
    """Get the global ToolManager instance."""
    return ToolManager()


# Convenience functions
async def run_tool(
    tool_name: str,
    command: str,
    timeout: float = 300,
    **kwargs
) -> ToolExecution:
    """Run a tool with the global ToolManager."""
    return await get_tool_manager().execute(
        tool_name, command, timeout=timeout, **kwargs
    )


async def kill_tool(tool_id: str) -> bool:
    """Kill a tool by ID."""
    return await get_tool_manager().kill_tool(tool_id)



import uuid
from datetime import UTC, datetime
from typing import Any

from pydantic import BaseModel, Field


def _generate_agent_id() -> str:
    return f"agent_{uuid.uuid4().hex[:8]}"


class AgentState(BaseModel):
    agent_id: str = Field(default_factory=_generate_agent_id)
    agent_name: str = "ExaaiAgnt"
    parent_id: str | None = None
    sandbox_id: str | None = None
    sandbox_token: str | None = None
    sandbox_info: dict[str, Any] | None = None

    task: str = ""
    iteration: int = 0
    max_iterations: int = 300
    completed: bool = False
    stop_requested: bool = False
    waiting_for_input: bool = False
    llm_failed: bool = False
    waiting_start_time: datetime | None = None
    final_result: dict[str, Any] | None = None
    max_iterations_warning_sent: bool = False

    messages: list[dict[str, Any]] = Field(default_factory=list)
    context: dict[str, Any] = Field(default_factory=dict)

    start_time: str = Field(default_factory=lambda: datetime.now(UTC).isoformat())
    last_updated: str = Field(default_factory=lambda: datetime.now(UTC).isoformat())

    actions_taken: list[dict[str, Any]] = Field(default_factory=list)
    observations: list[dict[str, Any]] = Field(default_factory=list)

    errors: list[str] = Field(default_factory=list)

    def increment_iteration(self) -> None:
        self.iteration += 1
        self.last_updated = datetime.now(UTC).isoformat()

    def add_message(self, role: str, content: Any) -> None:
        self.messages.append({"role": role, "content": content})
        self.last_updated = datetime.now(UTC).isoformat()

    def add_action(self, action: dict[str, Any]) -> None:
        self.actions_taken.append(
            {
                "iteration": self.iteration,
                "timestamp": datetime.now(UTC).isoformat(),
                "action": action,
            }
        )

    def add_observation(self, observation: dict[str, Any]) -> None:
        self.observations.append(
            {
                "iteration": self.iteration,
                "timestamp": datetime.now(UTC).isoformat(),
                "observation": observation,
            }
        )

    def add_error(self, error: str) -> None:
        self.errors.append(f"Iteration {self.iteration}: {error}")
        self.last_updated = datetime.now(UTC).isoformat()

    def update_context(self, key: str, value: Any) -> None:
        self.context[key] = value
        self.last_updated = datetime.now(UTC).isoformat()

    def set_completed(self, final_result: dict[str, Any] | None = None) -> None:
        self.completed = True
        self.final_result = final_result
        self.last_updated = datetime.now(UTC).isoformat()

    def request_stop(self) -> None:
        self.stop_requested = True
        self.last_updated = datetime.now(UTC).isoformat()

    def should_stop(self) -> bool:
        return self.stop_requested or self.completed or self.has_reached_max_iterations()

    def is_waiting_for_input(self) -> bool:
        return self.waiting_for_input

    def enter_waiting_state(self, llm_failed: bool = False) -> None:
        self.waiting_for_input = True
        self.waiting_start_time = datetime.now(UTC)
        self.llm_failed = llm_failed
        self.last_updated = datetime.now(UTC).isoformat()

    def resume_from_waiting(self, new_task: str | None = None) -> None:
        self.waiting_for_input = False
        self.waiting_start_time = None
        self.stop_requested = False
        self.completed = False
        self.llm_failed = False
        if new_task:
            self.task = new_task
        self.last_updated = datetime.now(UTC).isoformat()

    def has_reached_max_iterations(self) -> bool:
        return self.iteration >= self.max_iterations

    def is_approaching_max_iterations(self, threshold: float = 0.85) -> bool:
        return self.iteration >= int(self.max_iterations * threshold)

    def has_waiting_timeout(self) -> bool:
        if not self.waiting_for_input or not self.waiting_start_time:
            return False

        if (
            self.stop_requested
            or self.llm_failed
            or self.completed
            or self.has_reached_max_iterations()
        ):
            return False

        elapsed = (datetime.now(UTC) - self.waiting_start_time).total_seconds()
        return elapsed > 600

    def has_empty_last_messages(self, count: int = 3) -> bool:
        if len(self.messages) < count:
            return False

        last_messages = self.messages[-count:]

        for message in last_messages:
            content = message.get("content", "")
            if isinstance(content, str) and content.strip():
                return False

        return True

    def get_conversation_history(self) -> list[dict[str, Any]]:
        return self.messages

    def get_execution_summary(self) -> dict[str, Any]:
        return {
            "agent_id": self.agent_id,
            "agent_name": self.agent_name,
            "parent_id": self.parent_id,
            "sandbox_id": self.sandbox_id,
            "sandbox_info": self.sandbox_info,
            "task": self.task,
            "iteration": self.iteration,
            "max_iterations": self.max_iterations,
            "completed": self.completed,
            "final_result": self.final_result,
            "start_time": self.start_time,
            "last_updated": self.last_updated,
            "total_actions": len(self.actions_taken),
            "total_observations": len(self.observations),
            "total_errors": len(self.errors),
            "has_errors": len(self.errors) > 0,
            "max_iterations_reached": self.has_reached_max_iterations() and not self.completed,
        }



from typing import Any, ClassVar

from textual.widgets import Static

from .base_renderer import BaseToolRenderer
from .registry import register_tool_renderer


@register_tool_renderer
class WebSearchRenderer(BaseToolRenderer):
    tool_name: ClassVar[str] = "web_search"
    css_classes: ClassVar[list[str]] = ["tool-call", "web-search-tool"]

    @classmethod
    def render(cls, tool_data: dict[str, Any]) -> Static:
        args = tool_data.get("args", {})
        query = args.get("query", "")

        header = "🌐 [bold #60a5fa]Searching the web...[/]"

        if query:
            query_display = query[:100] + "..." if len(query) > 100 else query
            content_text = f"{header}\n  [dim]{cls.escape_markup(query_display)}[/]"
        else:
            content_text = f"{header}"

        css_classes = cls.get_css_classes("completed")
        return Static(content_text, classes=css_classes)



import os

from .executor import (
    execute_tool,
    execute_tool_invocation,
    execute_tool_with_validation,
    extract_screenshot_from_result,
    process_tool_invocations,
    remove_screenshot_from_result,
    validate_tool_availability,
)
from .k8s_scanner import (
    K8sScanner,
    check_pod_security,
    check_rbac,
    scan_cluster,
)
from .prompt_injection import (
    PromptInjectionScanner,
    analyze_llm_response,
    detect_jailbreak_success,
    generate_injection_payloads,
    scan_for_prompt_injection,
)
from .registry import (
    ImplementedInClientSideOnlyError,
    get_tool_by_name,
    get_tool_names,
    get_tools_prompt,
    needs_agent_state,
    register_tool,
    tools,
)
from .response_analyzer import (
    Detection,
    DetectionType,
    ResponseAnalyzer,
    analyze_response,
    get_response_analyzer,
)
from .smart_fuzzer import (
    FuzzPayload,
    ParamType,
    SmartFuzzer,
    VulnCategory,
    fuzz_parameter,
    get_smart_fuzzer,
)
from .tool_prompts import (
    get_all_tool_prompts,
    get_analyzer_prompt,
    get_fuzzer_prompt,
    get_security_testing_prompt,
    get_validator_prompt,
    get_waf_bypass_prompt,
)
from .vuln_validator import (
    Severity,
    VulnerabilityReport,
    VulnStatus,
    VulnValidator,
    create_vuln_report,
    get_vuln_validator,
)
from .waf_bypass import (
    WAFBypass,
    WAFDetector,
    WAFType,
    detect_waf,
    generate_bypasses,
    get_waf_bypass,
)


SANDBOX_MODE = os.getenv("EXAAI_SANDBOX_MODE", "false").lower() == "true"

HAS_PERPLEXITY_API = bool(os.getenv("PERPLEXITY_API_KEY"))

if not SANDBOX_MODE:
    from .agents_graph import *  # noqa: F403
    from .browser import *  # noqa: F403
    from .file_edit import *  # noqa: F403
    from .finish import *  # noqa: F403
    from .notes import *  # noqa: F403
    from .proxy import *  # noqa: F403
    from .python import *  # noqa: F403
    from .reporting import *  # noqa: F403
    from .terminal import *  # noqa: F403
    from .thinking import *  # noqa: F403

    if HAS_PERPLEXITY_API:
        from .web_search import *  # noqa: F403
else:
    from .browser import *  # noqa: F403
    from .file_edit import *  # noqa: F403
    from .notes import *  # noqa: F403
    from .proxy import *  # noqa: F403
    from .python import *  # noqa: F403
    from .terminal import *  # noqa: F403

__all__ = [
    "ImplementedInClientSideOnlyError",
    "execute_tool",
    "execute_tool_invocation",
    "execute_tool_with_validation",
    "extract_screenshot_from_result",
    "get_tool_by_name",
    "get_tool_names",
    "get_tools_prompt",
    "needs_agent_state",
    "process_tool_invocations",
    "register_tool",
    "remove_screenshot_from_result",
    "tools",
    "validate_tool_availability",
    # WAF Bypass
    "get_waf_bypass",
    "WAFBypass",
    "WAFDetector",
    "WAFType",
    "detect_waf",
    "generate_bypasses",
    # Smart Fuzzer
    "get_smart_fuzzer",
    "SmartFuzzer",
    "ParamType",
    "VulnCategory",
    "FuzzPayload",
    "fuzz_parameter",
    # Response Analyzer
    "get_response_analyzer",
    "ResponseAnalyzer",
    "DetectionType",
    "Detection",
    "analyze_response",
    # Vulnerability Validator
    "get_vuln_validator",
    "VulnValidator",
    "VulnStatus",
    "Severity",
    "VulnerabilityReport",
    "create_vuln_report",
    # Prompt Injection Scanner
    "PromptInjectionScanner",
    "scan_for_prompt_injection",
    "generate_injection_payloads",
    "analyze_llm_response",
    "detect_jailbreak_success",
    # K8s Scanner
    "K8sScanner",
    "scan_cluster",
    "check_rbac",
    "check_pod_security",
    # Tool Prompts
    "get_fuzzer_prompt",
    "get_analyzer_prompt",
    "get_validator_prompt",
    "get_waf_bypass_prompt",
    "get_security_testing_prompt",
    "get_all_tool_prompts",
]



"""
AI Prompt Injection Scanner - Detect LLM vulnerabilities in AI applications.

This module provides tools for testing AI/LLM applications for:
- Direct prompt injection
- Indirect prompt injection
- Jailbreak attempts
- Data exfiltration via prompts
- System prompt leakage
"""

from .prompt_injection_actions import (
    PromptInjectionScanner,
    analyze_llm_response,
    detect_jailbreak_success,
    generate_injection_payloads,
    scan_for_prompt_injection,
)


__all__ = [
    "PromptInjectionScanner",
    "analyze_llm_response",
    "detect_jailbreak_success",
    "generate_injection_payloads",
    "scan_for_prompt_injection",
]



from abc import ABC, abstractmethod
from typing import TypedDict


class SandboxInfo(TypedDict):
    workspace_id: str
    api_url: str
    auth_token: str | None
    tool_server_port: int
    agent_id: str


class AbstractRuntime(ABC):
    @abstractmethod
    async def create_sandbox(
        self,
        agent_id: str,
        existing_token: str | None = None,
        local_sources: list[dict[str, str]] | None = None,
    ) -> SandboxInfo:
        raise NotImplementedError

    @abstractmethod
    async def get_sandbox_url(self, container_id: str, port: int) -> str:
        raise NotImplementedError

    @abstractmethod
    async def destroy_sandbox(self, container_id: str) -> None:
        raise NotImplementedError



from typing import Any, Literal

from exaaiagnt.tools.registry import register_tool

from .python_manager import get_python_session_manager


PythonAction = Literal["new_session", "execute", "close", "list_sessions"]


@register_tool
def python_action(
    action: PythonAction,
    code: str | None = None,
    timeout: int = 30,
    session_id: str | None = None,
) -> dict[str, Any]:
    def _validate_code(action_name: str, code: str | None) -> None:
        if not code:
            raise ValueError(f"code parameter is required for {action_name} action")

    def _validate_action(action_name: str) -> None:
        raise ValueError(f"Unknown action: {action_name}")

    manager = get_python_session_manager()

    try:
        match action:
            case "new_session":
                return manager.create_session(session_id, code, timeout)

            case "execute":
                _validate_code(action, code)
                assert code is not None
                return manager.execute_code(session_id, code, timeout)

            case "close":
                return manager.close_session(session_id)

            case "list_sessions":
                return manager.list_sessions()

            case _:
                _validate_action(action)  # type: ignore[unreachable]

    except (ValueError, RuntimeError) as e:
        return {"stderr": str(e), "session_id": session_id, "stdout": "", "is_running": False}



import io
import re
import signal
import sys
import threading
from typing import Any

from IPython.core.interactiveshell import InteractiveShell


MAX_STDOUT_LENGTH = 10_000
MAX_STDERR_LENGTH = 5_000


class PythonInstance:
    def __init__(self, session_id: str) -> None:
        self.session_id = session_id
        self.is_running = True
        self._execution_lock = threading.Lock()

        import os

        # Use configurable workspace path with fallback
        workspace = os.getenv("EXAAI_WORKSPACE", "/workspace")
        if os.path.isdir(workspace):
            os.chdir(workspace)

        # Add current directory and workspace to sys.path to ensure local imports work
        sys.path.insert(0, os.getcwd())
        if workspace not in sys.path:
            sys.path.insert(0, workspace)

        self.shell = InteractiveShell()
        self.shell.init_completer()
        self.shell.init_history()
        self.shell.init_logger()

        self._setup_proxy_functions()

    def _setup_proxy_functions(self) -> None:
        try:
            from exaaiagnt.tools.proxy import proxy_actions

            proxy_functions = [
                "list_requests",
                "list_sitemap",
                "repeat_request",
                "scope_rules",
                "send_request",
                "view_request",
                "view_sitemap_entry",
            ]

            proxy_dict = {name: getattr(proxy_actions, name) for name in proxy_functions}
            self.shell.user_ns.update(proxy_dict)
        except ImportError:
            # Proxy functions are optional, only available in full runtime
            pass

    def _validate_session(self) -> dict[str, Any] | None:
        if not self.is_running:
            return {
                "session_id": self.session_id,
                "stdout": "",
                "stderr": "Session is not running",
                "result": None,
            }
        return None

    def _setup_execution_environment(self, timeout: int) -> tuple[Any, io.StringIO, io.StringIO]:
        stdout_capture = io.StringIO()
        stderr_capture = io.StringIO()

        def timeout_handler(signum: int, frame: Any) -> None:
            raise TimeoutError(f"Code execution timed out after {timeout} seconds")

        old_handler = signal.signal(signal.SIGALRM, timeout_handler)
        signal.alarm(timeout)

        sys.stdout = stdout_capture
        sys.stderr = stderr_capture

        return old_handler, stdout_capture, stderr_capture

    def _cleanup_execution_environment(
        self, old_handler: Any, old_stdout: Any, old_stderr: Any
    ) -> None:
        signal.signal(signal.SIGALRM, old_handler)
        sys.stdout = old_stdout
        sys.stderr = old_stderr

    def _truncate_output(self, content: str, max_length: int, suffix: str) -> str:
        if len(content) > max_length:
            return content[:max_length] + suffix
        return content

    def _format_execution_result(
        self, execution_result: Any, stdout_content: str, stderr_content: str
    ) -> dict[str, Any]:
        # Strip ANSI codes from output
        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        stdout_content = ansi_escape.sub("", stdout_content)
        stderr_content = ansi_escape.sub("", stderr_content)

        stdout = self._truncate_output(
            stdout_content, MAX_STDOUT_LENGTH, "... [stdout truncated at 10k chars]"
        )

        if execution_result.result is not None:
            if stdout and not stdout.endswith("\n"):
                stdout += "\n"
            result_repr = repr(execution_result.result)
            result_repr = self._truncate_output(
                result_repr, MAX_STDOUT_LENGTH, "... [result truncated at 10k chars]"
            )
            stdout += result_repr

        stdout = self._truncate_output(
            stdout, MAX_STDOUT_LENGTH, "... [output truncated at 10k chars]"
        )

        stderr_content = stderr_content if stderr_content else ""
        stderr_content = self._truncate_output(
            stderr_content, MAX_STDERR_LENGTH, "... [stderr truncated at 5k chars]"
        )

        if (
            execution_result.error_before_exec or execution_result.error_in_exec
        ) and not stderr_content:
            stderr_content = "Execution error occurred"

        return {
            "session_id": self.session_id,
            "stdout": stdout,
            "stderr": stderr_content,
            "result": repr(execution_result.result)
            if execution_result.result is not None
            else None,
        }

    def _handle_execution_error(self, error: BaseException) -> dict[str, Any]:
        error_msg = str(error)
        error_msg = self._truncate_output(
            error_msg, MAX_STDERR_LENGTH, "... [error truncated at 5k chars]"
        )

        return {
            "session_id": self.session_id,
            "stdout": "",
            "stderr": error_msg,
            "result": None,
        }

    def execute_code(self, code: str, timeout: int = 30) -> dict[str, Any]:
        session_error = self._validate_session()
        if session_error:
            return session_error

        with self._execution_lock:
            old_stdout, old_stderr = sys.stdout, sys.stderr

            try:
                old_handler, stdout_capture, stderr_capture = self._setup_execution_environment(
                    timeout
                )

                try:
                    execution_result = self.shell.run_cell(code, silent=False, store_history=True)
                    signal.alarm(0)

                    return self._format_execution_result(
                        execution_result, stdout_capture.getvalue(), stderr_capture.getvalue()
                    )

                except (TimeoutError, KeyboardInterrupt, SystemExit) as e:
                    signal.alarm(0)
                    return self._handle_execution_error(e)

            finally:
                self._cleanup_execution_environment(old_handler, old_stdout, old_stderr)

    def close(self) -> None:
        self.is_running = False
        self.shell.reset(new_session=False)

    def is_alive(self) -> bool:
        return self.is_running
